/*++++++++++++++
.IDENTIFICATION aserver.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT
.KEYWORDS       Internet Client/Server, run as a Daemon
.VERSION  1.0   23-Jul-1992
.VERSION  1.1   17-Aug-1992: possibility of File Exchange
.VERSION  1.2   26-Aug-1992: re-arrangement of Configuration File
.VERSION  1.3   19-Sep-1992: added .init
.VERSION  1.4   21-Oct-1992: Command may be preceded by a - to
			eliminate the stdout in pipes.
.VERSION  1.5   12-Dec-1992: Added -bsk option
.VERSION  2.0   10-May-1993: Added -a option (authorize file)
.VERSION  2.1   16-Jul-1993: Added ! to save errors in temporary file
.VERSION  2.2   29-Jul-1993: Added -launch option
.VERSION  2.3   20-Aug-1993: Added function rm0  (Remove Function)
.VERSION  2.4   01-Sep-1993: Check security, close properly quote(s)
			to shell (executed in shell(...) routine)
.VERSION  2.5   28-Jan-1994: Allow the usage of /etc/passwd
.VERSION  2.6   03-Mar-1994: USER terminated by ! indicates root privilege
.VERSION  2.7   21-Apr-1995: Added PID Env. Variable
.VERSION  2.8   18-Dec-1998: Adapted for Linux
.VERSION  2.9   30-Oct-1999: startlog possibility
.VERSION  2.91  08-Dec-1999: allow several actions (e.g. sauvelog)
.VERSION  3.0   14-Jan-2000: Writes also /var/log/xferlog if
				"[" is in the logged text
.VERSION  3.1   30-Jan-2001: Write a log file if logged text starts by @
.VERSION  3.2   20-Jan-2003: Write a file ==> set its size to zero.
.VERSION  3.3   05-Feb-2003: fork+_exit when -launch option; close log
	Accept Hangup signal, and verifies file change every CHECKFILE_INTERVAL
.VERSION  3.3a  08-Oct-2003: Use wait3 to gobble all waiting children.
.VERSION  3.4   17-Oct-2003: Was exiting in case of bad acknowledgement...
.VERSION  3.5   04-Nov-2003: Limits on size e.g. *write_1M>/tmp
				to write up to 1Mb
.VERSION  3.6   04-Mar-2004: Duration function to change the default
				limit time allowed to a child.
.VERSION  3.7   01-Dec-2004: Add "save" in sk_file
				and the regular Cleanup
.VERSION  3.8   18-Dec-2004: "save" function: return the file name.
.VERSION  3.9   22-Oct-2005: Accept bz2 compressions.
.VERSION  4.0   01-Mar-2006: Accept duration in hours;
			message in case of time-out and exit 2
.VERSION  4.1   24-Mar-2006: Change a bit the options...
.VERSION  4.2   26-Sep-2006: Add the 'x' option + |-command
.VERSION  4.3   16-Jan-2007: Cleanup in a function
.VERSION  4.4   10-Feb-2007: 64-bit
.VERSION  4.5   01-Mar-2007: Tested on 64-bit
.VERSION  4.6   05-Sep-2007: Extended services: add put / get
.VERSION  4.7   09-Oct-2007: Take care of end backslashes
.VERSION  4.71  23-Oct-2007: Pb with empty lines
.VERSION  4.8   07-Dec-2007: Remove also temporary files (starting by t or T)

.COMMENTS       Listens to a service port, and executes shell commands
	from a provided list, or read/write files.
	The protocol for file exchange is:
	Server sends ^F:  client has to send a buffer
	Server sends ^B:  client will get Buff mode.

	Buff mode consists in sending a 4-byte buffer with length (in NET
	int format). The buffer is then sent.

	The configuration File is made of colon-separated columns
	containing:
	1) The mnemonic (what is sent by the Client)
	2) The translation; a command preceded by a
	       * is defined here.
	       - the contents of stderr is ignored
	       ! the stderror is redirected to a file
	       + turn on verbose mode
	       | a command which reads the socket
	3) The users allowed to execute the operation.
	4) The text logged for each call
	5) The list of the possible parameters
	6) Explanations
---------------*/

#define VERSION "4.8"

#ifndef CHECKFILE_INTERVAL	/* Verification of file config change */
#define CHECKFILE_INTERVAL  900	/* ... each 15min */
#endif

#ifndef TIME_MAX		/* Max time allowed to a child process */
#define TIME_MAX 	3600
#endif

#include <skserv.h>	/* local.  */
#include <signal.h>	/* SIGCHLD */
#include <syslog.h>	/* Levels of Logging */
#include <stdlib.h>	/* qsort, bserach, getenv */
#include <unistd.h>	/* alarm, lseek, etc */
#include <errno.h>	/* system error messages	*/
#include <string.h>
#include <setjmp.h>	/* To exit from Interrupt Handler */
#include <stdio.h>
#include <fcntl.h>	/* File mnemonics */
#include <ctype.h>	/* Character classes */
#include <time.h>	/* To compute times  */
#include <dirent.h>	/* Reading directorties */
#include <sys/wait.h>
#include <sys/stat.h>	/* Find file params  */
#include <sys/types.h>
#ifndef O_BINARY
#define O_BINARY 0
#endif
char *sk_caller();
#define bzero(a,n)	memset(a,0,n)
#define isendof(c)	(isspace(c)||(!(c)))

#ifndef _ARGS
#ifdef __STDC__
#define _ARGS(A)      A   /* For ANSI, define as    A */
#else
#define _ARGS(A)      ()  /* For ANSI, define as    A */
#endif
#endif
#ifndef LEN_CWD
#define LEN_CWD	512	  /* Max length of file names */
#endif

typedef int (*SORT_FCT)_ARGS((const void *, const void *)) ;

static FILE *logfile;	  /* Logfile in daemon mode   */
static char *logname;	  /* Name of log file ---     */
static char *logname_0;	  /* Moved log file (rotlog)  */
extern int errno;
static void onHangup(), periodic_check();

#define ITEMS(t)	(sizeof(t)/sizeof(t[0]))
#ifdef MAX
#undef MAX
#endif
#define MAX(x,y)	((x) > (y) ? x : y)

/* xferlog options */
static char *xferlog_options[] = {
   "z",        "z",
   "Z",        "C",
   "bz2",      "j",
   "zip",      "K",
   "tar",      "T",
   "tar.gz",   "Tz",
   "tar.bz2",  "Tj",
   "tar.Z",    "TC",
   "fit",      "F",
   "fits",     "F",
   "fit.Z",    "FC",
   "fits.Z",   "FC",
   "fit.bz2",  "Fj",
   "fits.bz2", "Fj",
   "fits.z",   "Fz",
   "fits.gz",  "Fz",
   "fit.gz",   "Fz",
   "fit.z",    "Fz",
   "fgz",      "Fz",
   "fiZ",      "FC",
 (char *)0 } ;

static int duration_suffix[] = {	/* V4.0 */
  's', 1,
  'm', 60,
  'h', 3600,
  'd', 86400,
  'j', 86400,
  'w', 7*3600,
  0
};

typedef struct {	/* List of Available Commands */
	int count;
	int bytes;	/* Length of text */
	char **table;	/* count * 6	*/
	char *text;	/* length bytes	*/
  } COMS;

typedef int (*iFCT)();
typedef struct {
	char *name;	/* The name as it appears in the Configuration File */
	iFCT  action;	/* The function */
  } ACTION;

static COMS *commands = (COMS *)0;	/* Loaded Commands */

static char *ffile;	/* Configuratiion File */
static time_t ffile_time; /* Date of config file */
static int pid_root;	/* PID of root process */
static char *afile;	/* Authorization  File */
static char *sk_filename ;	/* Last opened tmpfile	   */
static char **thecom;	/* Set by trcom to current command */
static int bytes_transfered ;	/* Used by shell sk_file   */
static int max_transfer=0;	/* Limit number of bytes   */
static char *thecaller;	/* Who is on the other side...	   */
static char *com_err;	/* Set by trcom   		   */
static int  com_errno;
static char *acombuf;	/* Command buffer */
static int   lcombuf;	/* Length of ...  */
static char eof;	/* End indicator  */
static char ov;		/* Verbose option */
static int interrupted;	/* Set by OnIntr  */
static jmp_buf jmp_env;
static int toclose;	/* File to close in case of Interrupt */
static int jour;	/* Current day (to rotate logs)	*/

#if 0
/* Cleanup saved files */
static char cleanup_command[] = "\
find $HOME -name U\\* -print | gawk -F/ 'BEGIN{now=systime()}\
{ if(substr($NF,1,1) != \"U\") next;\
  s = substr($NF,2)+0; \
  if(s<1000000) next;\
  if (s>now) next;\
  print \"rm\", $0;\
  system(\"/bin/rm \" $0)\
}' 2>&1";
#endif

static char usage[] = "\
Usage: %s [-inet|-launch] [-uid] [-v[v]] [-bsk sksize] [-log logfile] \n\
       [-p pidfile] [-f file] -D [dir] [-x cmd] [-c conx] [-t timeout] \n\
       [-a pwfile|-A] service\n";
static char pgm[] = "aserver";
static char help[] = "\
    -inetd: start the program by inetd (socket already open)\n\
   -launch: forks for each connection request\n\
      -uid: execute setuid(uid number)\n\
        -v: Verbose option\n\
       -vv: Very Verbose option\n\
 -bsk blks: specify the physical socket size (default $SK_bsk)\n\
 -log file: File where to log details\n\
   -p file: File where to write out the pid (default; stderr)\n\
   -f file: List of available functions (Configuration File)\n\
   -D  dir: Directory where to save files (default: $HOME or /tmp)\n\
   -x comd: Additional command (-x + for extended set of commands)\n\
   -t maxs: timeout of one command (default=1h=3600s)\n\
        -A: Take information about Authorizaration in /etc/passwd\n\
   -a file: Authorization File\n\
   -c size: max number of pendent connection requests\n\
   service: service name or number\n\
The Functions are in a 6-column table (separator :) as:\n\
 nmemonic : command : list-of-users : text-for-log : parms : Explain\n\
";

/* Standard Functions, always available, locally defined */
/* WARNING: passwd MUST BE THE LAST in the list !!!!!!!  */
/* whoami:/usr/ucb/whoami::-::tell my name\n\ -- REMOVED (security) */

static char stdconf[] = "\
?:*help::-::Display help\n\
quit:*quit::QUIT asked::Terminate\n\
cd:*cd::CD :[directory_name]:Change directory\n\
setenv:*setenv::SETENV :symbol value:Define a symbol\n\
rm0:*rm0::-:command [command...]:Remove a Command\n\
duration:*duration::-:time(s):Give me the specified time for the execution\n\
echo:echo::-:text:echo the text\n\
ls:ls::-:[file-template]:list files\n\
pwd:pwd::-::print working directory\n\
printenv:/usr/bin/printenv::-::print known symbols\n\
startlog:*startlog::-::start logging onto /tmp/sk.log\n\
rotlog:*rotlog::-::rotate logs (only in daemon mode)\n\
passwd:*passwd::*changePASSW*:new_password:change password\n\
" ;

/* .init:*setenv HOME /tmp ; *cd /tmp\n\ */
static char more_conf[] = "\
save:*sk_file save :Name_or_Delay:-:[file]:Save a file to disk (in $HOME)\n\
putmp:*sk_file write_50M /tmp/t-$PID::-:file:Write a temporary file (<50Mb)\n\
rmtmp:rm -f /tmp/t-$PID::-::Remove file written by putmp\n\
get:*sk_file read::-:file:Read a file\n\
put:*sk_file write::-:file:Write a file\n\
cat:-cat::-::Display a file\n\
catar:-tar cf -::-::Create a TAR on stdout\n\
xtar:-|tar xf -::-::Write a directory structure from stdin\n\
catar.gz:-tar cfz -::-::Create a TAR.gz on stdout\n\
xtar.gz:-|tar xfz -::-::Write a directory structure from gzipped stdin\n\
dir:ls -l:::: list files with attributes (ls -l)\n\
";

#define mexp(b, l)	(b ? realloc(b,l) : malloc(l))	/* Reallocation */

/*============================================================================*/
static int strloc(text, c)
/*++++++++++++++++
.PURPOSE  Locate specified character
.RETURNS  Index of located char
-----------------*/
  char *text;     /* IN: String to interpret */
  int  c;         /* IN: Char to look for */
{
  char *s;
    for ( s = text; *s; s++)      if (*s == c)    break;
    return(s-text);
}

static int streloc(text, c)
/*++++++++++++++++
.PURPOSE  Locate non-escaped character
.RETURNS  Index of located char
-----------------*/
  char *text;     /* IN: String to interpret */
  int  c;         /* IN: Char to look for */
{
  char *s;
    if (*text == c) return(0);
    for ( s = text; *s; s++) {
	if (*s != c)    continue;
	if (s[-1] != '\\') break;	/* non-escaped */
    }
    return(s-text);
}

static char *trim(text, op)
/*++++++++++++++++
.PURPOSE  Remove Leading + Trailing Blanks
.RETURNS  Pointer to beginning of trimmed string
-----------------*/
  char *text;   /* MOD: String to "trim" */
  int  op;	/* IN: 1 for Head, 2 for Tail, 3 for both */
{
  char *p;
    if (op&2) for (p = text + strlen(text); --p >= text; ) {
	if (isspace(*p)) *p = 0;
	else break;
    }
    p = text;
    if (op&1) while (isspace(*p)) *(p++) = 0;
    return(p);
}
/*===========================================================================*/
static void error(text1,text2)
/*++++++++++++++++
.PURPOSE  Error exit
.RETURNS  -
-----------------*/
  char *text1, *text2;     /* IN: Text before error */
{
    sk_log(LOG_ERR, "%s: %s%s\n", pgm, text1, text2);
    sk_log(LOG_INFO, usage, pgm);
    exit(1);
}

static int send_errors(plug, efile)
/*++++++++++++++++
.PURPOSE  Returns the file with errors to the Client
.RETURNS  Number of errors; errfile is deleted
-----------------*/
  int  plug;	/* IN: Socket number for Output */
  char *efile;	/* IN: Command to execute */
{
  int f, n, nrec;
  char buf[BUFSIZ], *p, *e;

    nrec = 0;
    f = open(efile, 0);
    if (f<0) return(0);
    if (lseek(f, 0L, 2) > 1) {
	lseek(f, 0L, 0);
	p = buf; e = buf+sizeof(buf)-1;
	while ((n = read(f, p, e-p)) > 0) {
	    p += n; *(p--) = 0;
	    while ((p>=buf) && (*p != '\n')) --p;
	    if (p <= buf) p = (char *)0;
	    else *(p++) = 0;
	    sk_errsend(plug, buf);
	    if (p) strcpy(buf, p);
	    else buf[0] = 0;
	    p = buf + strlen(buf);
	    nrec++;
	}
    }
    close(f);
    unlink(efile);
    return(nrec);
}

static char *afind(table, item)
/*++++++++++++++++
.PURPOSE  Find the equivalence in table
.RETURNS  The equivalence / NULL
-----------------*/
  char **table ;	/* IN; Table with items */
  char *item;		/* IN: Name to look at  */
{
  char **a, *t, *i ;
    /* Is OK is words agree */
    for (a=table; *a; a += 2) {
	for (t = *a, i = item; *t && (*t == *i); t++, i++) ;
	if (isendof(*t) && isendof(*i)) return(a[1]) ;
    }
    return(*a) ;
}

static char *sname(name)
/*++++++++++++++++
.PURPOSE  Substitute env. variables in "name"
.RETURNS  Interpreted string
.REMARKS  Impremented via `echo name`
-----------------*/
  char *name;	/* IN: Name which may include '$'	*/
{
  static char buf[256]; char *p;
  FILE *pop; int i;
    i = strloc(name, '$');
    if (!name[i]) return(name);		/* No $ ... */

    strcpy(buf, "echo ");
    i = strlen(name); if (i > (sizeof(buf)-8)) i = sizeof(buf)-8;
    p = buf + strlen(buf);
    strncpy(p, name, i); p[i] = 0;
    pop = popen(buf, "r");
    if (! pop) {
	sk_perror("***sname");
	return((char *)0);
    }
    fgets(buf, sizeof(buf), pop);
    pclose(pop);
    for (i = strlen(buf); --i >= 0; ) { if (isgraph(buf[i])) break; }
    buf[++i] = 0;
    return(buf);
}

static int create_file(dir, name)
/*++++++++++++++++
.PURPOSE  Create a file to save data (triggered by *sk_file save)
.RETURNS  Opened file descriptor
.REMARKS  Added V3.7.
	If name is numeric, it's assumed to be a delay
	and file name is copied to sk_filename and has the template
		$HOME/dir/UxxxxxxxxxxPxxxxxx-n
	where U is followed by Unix date of expiration date.
	If name is blank, exporiration of 1 week (7 days)
-----------------*/
  char *dir;	/* IN: Directory in which to write the file */
  char *name;	/* IN: Name  orDelaration delay	*/
{
  static char time_symb[] = { 's', 'm', 'h',  'j',   'd',   'w',    0 };
  static int  time_unit[] = { 1,   60, 3600, 86400, 86400, 7*86400, 1 } ;
  char buf[286];
  char *b, *e, *p;
  time_t expires;
  int seq, fd;

    expires = 0;
    b = buf; e = b + sizeof(buf)-22;	/* Give room for delay... */
    e[0] = e[1] = e[2] = e[3] = 0;
    /* Never a full path ! */
    if (*dir == '/') {	/* Need a full path */
	dir++;
    }
    p = isgraph(*dir) ?
	dir :		/* Use the directory given */
	getenv("HOME");	/* Use the HOME directory  */
    while (isgraph(*p)&&(b<e)) *(b++) = *(p++);
	*(b++) = '/';

    if (!name) name = "";
    if (!*name) name = "3600";		/* Default = 1hr */
    if (isdigit(*name)) {
	expires = atoi(name);
	/* Take units */
	for (p=name; isdigit(*p); p++) ;
	expires *= time_unit[strloc(time_symb, *p)];
	if (expires > 7*86400) 		/* Don't accept more that 1 week */
	    expires = 7*86400;
	expires += time((time_t *)0);
        sprintf(b, "U%010ldP%05d", expires, getpid());
	b += strlen(b);
    }
    else {
        strncpy(b, name, e-b);
	b += strlen(b);
    }

    if (ov) sk_log(LOG_INFO, "create_file(%s, %s) => %s\n", dir, name, buf);
    /* Tries to open the file */
    fd = open(buf, O_CREAT | O_WRONLY | O_EXCL, 0644);
    if (fd<0) for (seq=1; errno == EEXIST; seq++) {
        sprintf(b, "-%d", seq);
	fd = open(buf, O_CREAT | O_WRONLY | O_EXCL, 0644);
	if (fd >=0) break;
    }
    if (fd<0) sk_perror(buf);

    /* Save the filename used */
    if (sk_filename) free(sk_filename);
    sk_filename = strdup(buf);
    return(fd);
}

static int shell(cmd, opt, plug)
/*++++++++++++++++
.PURPOSE  Execute a shell command
.RETURNS  0 = success
.REMARKS  Option is concatenated to cmd. Redirections and Subshells
		are FORBIDDEN in cmd.
	bytes_transfered is set to the number of bytes copied.
-----------------*/
  char *cmd;	/* IN: Command to execute, may start by | */
  char *opt;	/* IN: Supplementary (redirection) Option */
  int  plug;	/* IN: Socket number for Output */
{
  static char *buf; static int lbuf = 0;
  static char popop[2] = "r";
  FILE *pop, *popen();
  char *p, quote;
  int len;

    bytes_transfered = 0 ;	/* Nothing through socket (yet) */

    /* len, length of buf, must be a multiple of 512 */
    len = (513 + strlen(cmd) + (opt ? strlen(opt)+6 : 0)) & (~511);
    if (len > lbuf) {
	lbuf = len;
	buf = mexp(buf, lbuf);
    }
    buf[0] = 0;
    if (opt) strcpy(buf, "( ");
    if (*cmd == '|') {	/* V4.2: shell commands read the plug */
	popop[0] = 'w';
	cmd++;
    }
    else popop[0] = 'r';
    strcat(buf, cmd);

    /* Remove possibilities of File Redirection, pipe, etc... */
    for (p=buf, quote=0; *p; p++) switch(*p) {
	case '\'':
	case '"':
	    if (!quote) quote = *p;
	    else if (quote == *p) quote = 0;
	    continue;
#if 0
	case '`': case '>': case '<': case '&': case '|': case '^':
	    if (quote != '\'' ) *p = '';
	    continue;
#endif
    }
    if (quote) *(p++) = quote, *p = 0;
    if (opt) sprintf(p, " ) %s", opt);
    if (ov) sk_log(LOG_INFO, "Pipe_%s <%s>\n", popop, buf);

    pop = popen(buf, popop);
    if (! pop) {
	sk_perror("***exec");
	com_err = "can't Exec: "; com_errno = errno;
	return(-1);
    }
    toclose = fileno(pop);	/* Tell file to close in INTERRUPT */
    if (setjmp(jmp_env) == 0) bytes_transfered = popop[0] == 'w' ?
	sk_fromclient(plug, toclose) :
	sk_toclient  (plug, toclose);
    else {			/* Retruns from  Interruption      */
	sk_putl(plug, 0L);
	sk_log(LOG_WARNING, "Interrupted returned to %d\n", __LINE__);
    }
    if(toclose) pclose(pop), toclose = 0;
    if (ov) sk_log(LOG_INFO, "Pipe terminated.\n", buf);
    if (bytes_transfered < 0) bytes_transfered = 0 ;
    return(0);
}

static int xferlog(time2, bytes, logo, prog)
/*++++++++++++++++
.PURPOSE  Log the transfer to the /var/log/xferlog file
.RETURNS  0 = success
.REMARKS
-----------------*/
  time_t time2[2]; 	/* IN: Begin / End of Transfer */
  int bytes ;		/* IN: Number of byutes transfered */
  char *logo ;		/* IN: Program logo, e.g. "[rCat]"	*/
  char *prog ;		/* IN: Program  + Options + Parameters  */
{
  static char logfile[] = "/var/log/xferlog" ;
  char buffer[300], *b, *o, *p, *e ;
  int fh, sec;

    /* sk_log(LOG_INFO, "Entered xferlog, log=%s\tprog=%s\n", logfile, prog); */
    bzero(buffer, sizeof(buffer)) ;
    strcpy(buffer, ctime(time2)) ;
    b = buffer + strloc(buffer, '\n') ; *(b++) = ' ' ;
    e = buffer + sizeof(buffer) -10 ;

    /* Transfer Time */
    sec = time2[1] - time2[0] ;
    if (sec <= 0) sec = 1 ;
    sprintf(b, "%d ", sec), b += strlen(b) ;

    /* Host from thecaller "francois@130.79.128.13(vizier)" */
    if ((p = strchr(thecaller, '@'))) p++ ;
    else p = "?no_node?" ;
    while (*p && (*p != ':') && (b<e)) *(b++) = *(p++) ;

    /* Bytes transfered */
    sprintf(b, " %d ", bytes), b += strlen(b) ;

    /* File name = Arguments to program (skipping the options) */
    p = prog + strlen(prog) - 1 ;
    while (isgraph(*p) && (b<e) &&  (p >= prog)) --p ; ++p ;
    if (!*p) p = prog ;
    while((b<e) &&  isgraph(*p)) *(b++) = *(p++) ;

    /* Binary Copy */
    strcpy(b, " b "), b += strlen(b) ;

    /* Options as in XFERLOG */
    p = prog;
    while (*p) {
	if (*p == '-') break ;
	while (isgraph(*p)) p++ ;
	while (isspace(*p)) p++ ;
    }
    if (*p == '-') {
	o = afind(xferlog_options, p+1) ;
	if (!o) o = "*", sk_log(LOG_ERR,
	    "option not in xferlog_options: %s\n", p) ;
    }
    else o = (char *)0 ;
    if (!o) o = "_" ;
    strcpy(b, o), b += strlen(b) ;

    /* Output, Anonymous */
    strcpy(b, " o a "), b += strlen(b) ;

    /* User who made the copy */
    for (p=thecaller; *p && (*p != '@') && (b<e); p++, b++) *b = *p ;

    /* Logo of copy method (no blank!), and Complete Copy */
    for (p=logo, *(b++) = ' '; isgraph(*p) && (b<e); p++, b++) *b = *p ;
    strcpy(b, " 0 * c\n"), b += strlen(b) ;

    fh = open(logfile, O_WRONLY|O_APPEND|O_CREAT, -1) ;
    if (fh < 0) {
	sk_log(LOG_ERR, "Can't open file: %s\n", logfile) ;
	return (-1) ;
    }
    lockf(fh, 1, -1L) ; 	/* Lock File */
    write(fh, buffer, strlen(buffer)) ;
    close(fh) ;

    return(0) ;
}

/*============================================================================*
 *		Routines dealing with the Configuration File
 *============================================================================*/
#if 0		/* Can be defined as a macro in string.h */
static char *strdup(s1)
/*++++++++++++++++
.PURPOSE  Duplicate a string
.RETURNS  Copy of s1 in a new piece of memory
-----------------*/
  char *s1;	/* IN: String to duplicate */
{
  char *new;
    new = malloc(1+strlen(s1));
    strcpy(new, s1);
    return(new);
}
#endif

static int fmatch(s, t)
/*++++++++++++++++
.PURPOSE  Check whether s starts like t.
.RETURNS  0 if mismatch
-----------------*/
  char *s;	/* IN: String to check */
  char *t;	/* IN: Template, * can match anything BUT dots */
{
#ifdef DEBUG
    fprintf(stderr, "....fmatch <%s> starts with <%s>\n", s, t);
#endif
    while (*t) {
	if (*(s++) != *(t++)) return(0);
    }
    return(1);
}

static int verify(u, list, sep, fct)
/*++++++++++++++++
.PURPOSE  Check if a user (User@Node) matches a list of athorized people.
.RETURNS  0 if mismatch
-----------------*/
  char *u;	/* IN: User@Node to check */
  char *list;	/* IN: List of authorized people, separated by commas */
  int sep;	/* IN: Separator character */
  int (*fct)();	/* IN: Test function 	*/
{
  int i, st;
  char x;

    if (!*list)	return(1);	/* Nothing = Everybody Allowed */
    for (st = 0; !st; ) {
	i = strloc(list, sep);
	x = list[i]; list[i] = 0;
	st = (*fct)(u, list);
	list += i;
	if ((*list = x)) list++;
	else break;
    }
    return(st);
}


static int tdiff(t1, t2)
/*++++++++++++++++
.PURPOSE  Compare two commands
.RETURNS  Difference
.REMARKS
-----------------*/
  char **t1, **t2;	/* IN: Strings to Compare */
{
    return(strcmp(*t1, *t2));
}

static char *load_file(file, atime)
/*++++++++++++++++
.PURPOSE  Load the Commands File
.RETURNS  Contents of the file
.REMARKS  Date of file modification saved in atime
-----------------*/
  char *file;		/* IN: File to load  */
  time_t *atime;	/* OUT: Time of file */
{
  int fno, bytes;
  char *text;

    fno = open(file, O_RDONLY | O_BINARY);
    if (fno <= 0)	{
	sk_log(LOG_ERR, "Can't load file %s\n", file);
	return((char *)0);
    }
    bytes = lseek(fno, 0L, 2); lseek(fno, 0L, 0);
    text = malloc(bytes+1);
    read(fno, text, bytes); text[bytes] = 0;
    if (atime) {  struct stat buf;
        fstat(fno, &buf);
	*atime = buf.st_mtime;
    }
    close(fno);

    return(text);
}

static int free_coms(COMS *icoms)
/*++++++++++++++++
.PURPOSE  Free an icoms
.RETURNS  0 (OK)
-----------------*/
{
    if (!icoms) return(0);
    if (icoms->text)  free(icoms->text);
    if (icoms->table) free(icoms->table);
    if (icoms->text || icoms->table) {
	icoms->text = (char *)0;
	icoms->table = (char **)0;
	free(icoms);
    }
    return(0);
}

static COMS *append_coms(COMS *icoms, char *more_com)
/*++++++++++++++++
.PURPOSE  Append (or rplace) commands in coms
.RETURNS  Pointer to updated structure.
.REMARKS
    	more_com consists in a 6-column table (separator :) as:
    nmemonic : command : list-of-users : text-for-log : parms : Explain
	with a few pecularities:
    - mnemonic starts with .(dot)	=> only 2 fields
    - command  starts with *		=> Internal command
    -  ""       ""     "   -		=> redirect stdout to /dev/null
    - text-for-log starts with *	=> don't log parameters
-----------------*/
{
  COMS coms, *acom;
  char **pt, **a;
  char *pc, *pe, *p;
  int  i, j, k;
  int lcom = 0;

    /* No need for password command ? */
    if ((!afile) && (p = strstr(stdconf, "passwd:")))
	*p = 0;

    bzero(&coms, sizeof(coms));

    /* Load first the File in Memory */
    if (!icoms) {		/* Defaut commands not yet included */
	icoms = append_coms(&coms, stdconf);
    }
    coms.bytes = strlen(more_com) + icoms->bytes;
    coms.text  = malloc(1+coms.bytes);
    if (icoms->text)
	memcpy(coms.text, icoms->text, icoms->bytes);
    pc = coms.text+icoms->bytes;	/* Where more_input stored */
    strcpy(pc, more_com);
    coms.text[coms.bytes] = 0;

    /* V4.7: remove ending backslahses first */
    for (pe=pc+strlen(pc), p=pc; p<pe; p++) {
	i = strloc(p, '\n');
	if (i==0) continue;	/* Empty line */
	p += i;
	if (p[-1] == '\\') { *--p = ' '; *++p = ' '; }
    }
    for (j=icoms->count; pc < pe; pc += 1+i) {
	i = strloc(pc, '\n');
	if (i == 0) continue;		/* Empty line..	(added V4.71)    */
	p = pc + i;
	/* if ((i) && (p[-1] == '\\')) { p[-1] = p[0] = ' '; continue; } */
	if (*pc != '#')  {
	    j++;
	    if (ov) { int len;
		fprintf(stderr, "#...Cmd#%-3d : ", j);
		len = i-1; if (len>40) len=40;
		fwrite(pc, 1, len, stderr);
		fputc('\n', stderr);
	    }
	}
    }
    if (j == coms.count)	/* No additional command! */
	return(icoms);
    coms.count = j;

    	/* Allocate the table, and copy preceding index */
    coms.table = (char **)malloc(coms.count*6*sizeof(char *));
    for (i=0, j=icoms->count*6, k=(coms.text-icoms->text); i<j; i++)
	coms.table[i] = icoms->table[i] + k;

    	/* Fill the table  */
    pt = coms.table + 6*icoms->count;
    for (pc = coms.text + icoms->bytes; pc < pe; pc += 1+i) {
	i = strloc(pc, '\n');
	if (i == 0) continue;		/* Empty line..	*/
	if (*pc == '#') continue;
	if (!*pc) 	continue;	/* Empty lines 	*/
	pc[i] = 0;			/* Remove \n	*/
	lcom = MAX(lcom, i);		/* Max command length */
	for (k = 0, p = pc; ++k <= 6 ; p += j, pt++ ) { /* 6 fields */
	    j = strloc(p, ':');
	    if ((*pc == '.') && (k == 2)) j = strlen(p);
	    if (p[j]) p[j++] = 0;
	    	/* Don't remove trailing blanks for field 4 (log text) */
	    *pt = trim(p, k == 4 ? 1 : 3);
	}
	if (ov) {
	    fprintf(stderr, "#...Cmd#%-3d = %s\n", (pt-coms.table)/6, pt[-6]);
	}
	/* Replace field if already known */
	if ((a = (char **)bsearch(&pc, coms.table,
		    icoms->count, 6*sizeof(char *), (SORT_FCT)tdiff))) {
	    pt -= 6; coms.count -= 1; coms.bytes -= (i+1) ;
	    for (k=0; k<6; k++) a[k] = pt[k];
	}
    }

    /* V4.7: Verify the counts are OK ! */
    if ((coms.table+6*coms.count) != pt) {
	fprintf(stderr, "****aserver: found %d services instead of %d!\n",
	    (pt-coms.table)/6, coms.count);
	fprintf(stderr, "    %s\n", 
	    "duplicate names ? re-run with -v (verbose) option");
    }

	/* Sort now the table, to allow usage of bsearch */
    qsort(coms.table, coms.count, 6*sizeof(char *), (SORT_FCT)tdiff);

	/* Create a new COMS */
    acom = (COMS *)malloc(sizeof(COMS));
    *acom = coms;
    if (lcom == 0) lcom = 80;
    lcom = (lcom + 511) & (~511);		/* Round up buffer size */
    if (lcom > lcombuf) {
	if (acombuf) free(acombuf);
	acombuf = malloc(lcombuf = lcom);
    }
    /* Free the input coms */
    free_coms(icoms);
    return(acom);
}

static COMS *load_coms(COMS *icoms, char *file)
/*++++++++++++++++
.PURPOSE  Load the Commands File
.RETURNS  Pointer to relevant Structure
.REMARKS
	stdconf added to list.
    	The Functions are in a 6-column table (separator :) as:
    nmemonic : command : list-of-users : text-for-log : parms : Explain
	with a few pecularities:
    - mnemonic starts with .(dot)	=> only 2 fields
    - command  starts with *		=> Internal command
    -  ""       ""     "   -		=> redirect stdout to /dev/null
    - text-for-log starts with *	=> don't log parameters
  Note that the static ffile_time is set.
-----------------*/
{
  COMS *pcom;
  char *text;

    /* Load first the File in Memory */
    text = load_file(file, &ffile_time);
    if (!text) return((COMS *)0);

    pcom = append_coms(icoms, text);
    free(text);
    return(pcom);
}

static char *trcom(coms, text, user)
/*++++++++++++++++
.PURPOSE  Translate the Command (first word in text)
.RETURNS  The translated command / NULL if not allowed
.REMARKS  thecom is set when found, and com_err if error found
-----------------*/
  COMS *coms;	/* IN: List of available commands */
  char *text;	/* IN: Text to translate (take aliases) */
  char *user;	/* IN: User name (to check )		*/
{
  char *buf, *bufe;
  static char bcom[40] ;
  static char *acom = bcom ;
  char *p, *q, x ;

    buf = acombuf; bufe = buf + lcombuf - 1;

	/* Retrieve first word */
    for (p = text; isgraph(*p); p++) ;
    strncpy(bcom, text, sizeof(bcom)-1) ;
    bcom[p-text] = 0 ;

	/* Retrieve alias from table, and append parameters */
    thecom = (char **)bsearch(&acom, coms->table,
	coms->count, 6*sizeof(char *), (SORT_FCT)tdiff);
    if (!thecom) {
	com_err =  "****Unknown Command: "; com_errno = 0;
	return((char *)0);
    }
    strcpy(buf, thecom[1]);
    if (*thecom[2]) {	/* No all users are allowed... */
    	if (!sk_umatch(user, thecom[2])) {
	    com_err =  "****Denied  Command: "; com_errno = 0;
	    return((char *)0); /* Check Username */
    	}
    }
	/* Check for the existence of quotes:
	   x contains the non-closed quote;
	   we've then to keep parameters attached to the quote
	*/
    for (q=buf, x=0; *q; q++) switch(*q) {
	case '\'':
	case '"':
	    if (!x) x = *q;
	    else if (x == *q) x = 0;
	    continue;
    }
    if (x) while(isspace(*p)) p++;

    strncpy(q, p, bufe-q);
    return(buf);
}

/*============================================================================*
 *		Routines Referred by a * in the Configuration File
 *============================================================================*/
static char *setuser(caller)
/*++++++++++++++++
.PURPOSE  Add in the Environment USER   NODE   HOST
.RETURNS  The environment string
-----------------*/
  char *caller;	/* IN: User@Node(Host)/Service */
{
  char *env, *p, *pe, *val[3];
  static char bid[20];
  static char *key[3] = { "USER", "NODE", "HOST" };
  int len, i, lval[3];

	/* Define $PID */
    sprintf(bid, "PID=%d", getpid());
    if (ov) sk_log(LOG_INFO, "putenv(%s)\n", bid);
    putenv(bid);

	/* Decompose the caller string */
    val[0] = caller; 		   lval[0] = strloc(val[0], '@');
    val[1] = val[0] + lval[0] + 1; lval[1] = strloc(val[1], '(');
    val[2] = val[1] + lval[1] + 1; lval[2] = strloc(val[2], ')');

    len = lval[0] + lval[1] + lval[2];	/* Required allocation memory */
    env = malloc(len+3*6);		/* 3 keys */

    for (pe = env, i = ITEMS(key); --i >= 0; pe = p) {
	p = pe;
	strcpy(p, key[i]), p += strlen(p); *(p++) = '=';
	strncpy(p, val[i], lval[i]), p += lval[i]; *(p++) = 0;
	if (ov) sk_log(LOG_INFO, "putenv(%s)\n", pe);
	putenv(pe);
    }

    return(env);
}

static ACTION *find_action(name, array, items)
/*++++++++++++++++
.PURPOSE  Retrieve an action from its name
.RETURNS  Action found / NULL
.REMARKS  com_err contains a text in case of error.
-----------------*/
  char *name;	/* IN: Action name to retrieve */
  ACTION *array;	/* IN: Array of actions to look at */
  int  items;	/* IN: Number of items in array    */
{
  ACTION *pa; int i; char *p, x;

	/* Define the name properly */

    if (*name == '*') name++;
    for (p = name; isgraph(*p); p++) ;
    x = *p; *p = 0;

    for (pa=array, i=items; --i >= 0; pa++) {
	if (strcmp(pa->name, name) == 0) break;
    }
    if (i < 0) {
	com_err = "****Unknown *action: "; com_errno = 0;
	pa = (ACTION *)0;
    }
    *p = x;
    return(pa);
}

static int sk_file(op, plug)
/*++++++++++++++++
.PURPOSE  Read/Write a file from/into a Socket
.RETURNS  Bytes transferred / -1 for Error
.REMARKS  Use Buffered Mode, i.e. transmit Length, then buffer
	Errors are put into com_err
	op contains the operation AND the list of available files, e.g.
	Write>/dev/>/tmp
	Write_1M>/dev/>/tmp
	which allows to write onto files starting with /dev/  or  /tmp/
-----------------*/
  char *op;	/* IN: Operation Read (write to socket) / Write .. */
  int  plug;	/* IN: Socket */
{
  static char shift_prefix[] = { 'K', 10, 'M', 20, 'B', 9, 0 };
  char *p, *file, *files, x;
  int fd, fs, stat, fop;

    bytes_transfered = max_transfer = 0 ;
    for (p = op; isalpha(*p); p++) ;
    if (*p == '_') p++;	/* A limit is supplied  */
    if (isdigit(*p)) {	/* A limit is supplied  */
       	max_transfer = atoi(p);
       	while (isdigit(*p)) p++;
       	if (isalpha(*p)) {
	    fs = strloc(shift_prefix, toupper(*p));
	    if (shift_prefix[fs]) max_transfer <<= shift_prefix[fs+1];
	    else sk_log(LOG_ERR, "Bad size indicator '%c' in: %s\n", *p, op);
	    sk_maxbytes(max_transfer);
	    p++;
        }
    }
    files = p;			/* e.g. files -> ">/tmp/"  */
    while (isgraph(*p)) p++;	/* End of files */
    while (isspace(*p)) p++;
    file = p; 			/* Second Word = File Name */

    switch(*op) {
      case 'A': case 'a': fs = 1;	/* Default = Standard Output */
	fop = O_APPEND | O_CREAT | O_WRONLY;	break;
      case 'P': case 'p': case 'W': case 'w': fs = 1;
	fop = O_CREAT | O_WRONLY | O_TRUNC;	break;
      case 'G': case 'g': case 'R': case 'r': fs = 0;
	fop = O_RDONLY; break;
      case 'S': case 's': 		/* V3.7: Save a file */
        fop = O_CREAT | O_WRONLY | O_EXCL;
	if (*files == '>') files++;
	fs = create_file(files, file);
	while (isgraph(*file)) file++;	/* Filename is used  */
	break;
      default:
	sk_log(LOG_ERR, "Bad operation <%s>\n", op);
	com_err = "****Bad operation: "; com_errno = 0;
	return(-1);
    }

	/* Check the available files */
    com_err = "****Error ";
    if (*file) {	/* Check if Allowd File */
	for (p = files; isgraph(*p); p++) ;
	x = *p; *p = 0;
	stat = *(files++);	/* stat contains the Separator Char */
	if (stat) stat = !verify(file, files, stat, fmatch);
	*p = x;
	if (stat) {
	    com_err = "****Can't access file "; com_errno = 0;
	    return(-1);
	}
	if (sk_filename) free(sk_filename) ;
	sk_filename = strdup(sname(file)) ;
	fd = open(sk_filename, fop | O_BINARY, 0644);
    }
    else    fd = fs;
    if (fd < 0) {
	sk_perror(op); com_errno = errno;
	return(-1);
    }
    stat = 0;	/* Returned Status */

    if (fop&O_WRONLY) {	/* Write file, i.e. READ socket */
	/* Added V3.7: Write the name of the File */
	if (fop&O_EXCL) /* "save" a file -- return its name! */
	    sk_puts(plug, sk_filename);
	stat = sk_fromclient(plug, fd);
	if (fop&O_EXCL) {
	  char *buf;	/* "save" a file -- return its name! */
	    buf = malloc(strlen(sk_filename) + 64);
	    sprintf(buf, "%s (%d bytes)", sk_filename, stat);
	    sk_puts(plug, buf);
	    free(buf);
	}
    }
    else stat = sk_toclient(plug, fd);
    if (stat > 0) bytes_transfered = stat ;
    if (fd > 2) {
	if (close(fd)) sk_perror(op);
    }

    return(stat);
}

static int sk_cd(op, plug)
/*++++++++++++++++
.PURPOSE  Change directory
.RETURNS  0 / -1
-----------------*/
  char *op;	/* IN: Directory name */
  int  plug;	/* IN: Socket */
{
    if (!*op)	op = getenv("HOME");
    if (chdir(op) < 0) 	{
	sk_perror("*** CD error");
	com_err = "**** "; com_errno = errno;
	return(-1);
    }
    return(0);
}

static int sk_rm0(op, plug)
/*++++++++++++++++
.PURPOSE  Remove a command from list
.RETURNS  Number of commands effectively removed
.REMARKS  Limit to 60 commands
-----------------*/
  char *op;	/* IN: Commands (separated by blanks) to remove */
  int  plug;	/* IN: Socket */
{
  COMS *coms;	/* IN: List of available commands */
  char **com; char *p, *b, c;
  char *torm[60]; int n, nrm;

    coms = commands;
    if (!coms) return(0);	/* Configuration not (yet) loaded... */

    nrm = 0;	/* Number of commands to remove  */
    for (p=op; nrm<ITEMS(torm); ) {	/* Scan the text */
	while (isspace(*p)) p++;
	for (b=p; isgraph(*b); b++) ;
	if (b == p) break;
	c = *b; *b = 0;
    	com = (char **)bsearch(&p, coms->table,
	    coms->count, 6*sizeof(char *), (SORT_FCT)tdiff);
	if (com) torm[nrm++] = *com;
	p = b;	/* For next item in list to look at */
	if ((*p = c)) p++;
    }
	/* Remove now the marked commands */
    if (nrm == 0) return(0);
    while (--nrm >= 0) *torm[nrm] = '\177'; 	/* Mark commands to remove */
    qsort(coms->table, coms->count, 6*sizeof(char *), (SORT_FCT)tdiff);
    for (nrm = (coms->count-1)*6, n=0; *coms->table[nrm] == '\177'; nrm -= 6)
	n++;
    coms->count -= n;

    return(n);
}

#if 1		/* V3.6 Addition */
static int duration_plug;
static int duration_period = TIME_MAX;
static void duration_expired(s)
/*++++++++++++++++
.PURPOSE  Just write an error message in case of too long child
.RETURNS  --
-----------------*/
  int s;	/* IN: Signal number */
{
  char buf[80];
  int i;
    sk_log(LOG_ERR, "TIMEOUT duration_expired(%ds)\n", duration_period);
    /* Terminate the buffers */
    for (i=256; --i>=0; ) sk_putl(duration_plug, 0L);
    sprintf(buf, "****Too long (%ds) -- use the 'duration' command\n",
      duration_period);
    sk_errsend(duration_plug, buf);
    sleep(1);
    /* V4.0: Remove temporary file */
    if (sk_filename && (strncmp(sk_filename, "/tmp/", 5) == 0)) {
	sk_log(LOG_WARNING, "removing: %s\n", sk_filename);
	unlink(sk_filename);
    }
    exit(2);
}

static int sk_duration(op, plug)
/*++++++++++++++++
.PURPOSE  Change the alarm time (Added in V3.6)
.RETURNS  The set alarm
-----------------*/
  char *op;	/* IN: variable value */
  int  plug;	/* IN: Socket */
{
  int i;
  char buf[64];
    while(isspace(*op)) op++;	/* Skip Leading blanks */
    duration_period = atoi(op);
    while(isdigit(*op)) op++;
    if (isalpha(*op)) {
	for (i=0; duration_suffix[i]; i += 2) {
	    if (*op == duration_suffix[i]) { i++; break; }
	}
	if (duration_suffix[i]) duration_period *= duration_suffix[i];
    }
    if (duration_period <= 0) duration_period = TIME_MAX;
    if (duration_period > 7*86400) duration_period = 7*86400;	/* 1 week */
    sprintf(buf, "....Duration set to: %ds", duration_period);
    signal(SIGALRM, duration_expired);
    duration_plug = plug;
    alarm(duration_period);
    sk_errsend(plug, buf);
    return(duration_period);
}
#endif		/* V3.6 Addition */

static int sk_setenv(op, plug)
/*++++++++++++++++
.PURPOSE  Set Variable in Environment
.RETURNS  0 / -1
-----------------*/
  char *op;	/* IN: variable value */
  int  plug;	/* IN: Socket */
{
  char *p, *e, *env;

    while(isspace(*op)) op++;	/* Skip Leading blanks */
    env = malloc(2+strlen(op));	/* Don't forget the =  */
    for (p = op, e = env; isgraph(*p); p++, e++)  *e = *p;
    while(isspace(*p)) p++;
    *(e++) = '=';
    while(isgraph(*p)) *(e++) = *(p++); *e = 0;
    if(ov) sk_log(LOG_INFO, "putenv(%s)\n", env);
    return(putenv(env));
}

static int sk_help(op, plug)
/*++++++++++++++++
.PURPOSE  HELP function
.RETURNS  0
-----------------*/
  char *op;	/* IN: Operation Read (write to socket) / Write .. */
  int  plug;	/* IN: Socket */
{
  char buf[BUFSIZ]; char *p, **pt; int i;

    sk_put(plug, "List of possibilities:\n");
    if (commands) for(i = commands->count, pt=commands->table;
      --i >= 0; pt += 6) {
	if (**pt == '.') continue;
	if (pt[5] == "") sprintf(buf, "%-8s = %s", pt[0], pt[1]);
	else {
	    sprintf(buf, "%s %s", pt[0], pt[4]);
	    for (p = buf + strlen(buf); p < (buf+30); p++) *p = ' ';
	    *(p++) = ':'; *(p++) = ' ';
	    strcpy(p, pt[5]);
     	}
	sk_puts(plug, buf);
    }
    else  sk_put(plug, ".... I don't know, try ....\n");
    return(0);
}

static int quit()
/*++++++++++++++++
.PURPOSE  QUIT function
.RETURNS  0
-----------------*/
{
    eof = 1; return(0);
}
static int startlog() { return sk_servlog(1); }

static int rotlog()
/*++++++++++++++++
.PURPOSE  Rotate the log
.RETURNS  Number of log files
.REMARKS  Requires logname_0
-----------------*/
{
  static char emove[] = ".0123456789ABCDEF";
  struct stat buf;
  char *fname, *fnew;
  int i, n=0, st, len;

    if (!logname_0) 	/* No log file... */
	return(0);

    /* Must be done by the Main Process ! */
    if (pid_root && (pid_root != getpid())) {
	kill(pid_root, SIGHUP);
	return(0);
    }

    /* Does the 'main' file exist ? */
    if (stat(logname_0, &buf))
	return(0);

    /* Create 2 file names for rotation */

    len = strlen(logname_0)-1;
    fname = malloc(2*len+4);
    fnew  = fname + (len+2);
    strcpy(fname, logname_0);
    strcpy(fnew , logname_0);
    
    /* How many files to move ? */
    for (i=1; emove[i]; i++) {
	fname[len] = emove[i];
	if (stat(fname, &buf)) break;
    }

    /* Remove last file */
    if (!emove[i]) {
	fname[len] = emove[--i];
	st = unlink(fname);
	if (st) sk_perror(fname);
    }

    /* Rename all files */
    while (i>0) {
	fnew[len] = emove[i];
	fname[len] = emove[--i];
	st = rename(fname, fnew);
	if (st) 
	    sk_perror(fname);
	n++;
    }

    /* Free allocated memory and returns */
    free(fname);
    return(n);
}

static ACTION actions[] = {
	{ "help", 	sk_help},
	{ "quit", 	quit},
	{ "cd", 	sk_cd },
	{ "setenv", 	sk_setenv },
	{ "duration", 	sk_duration },
	{ "sk_file", 	sk_file },
	/* { "passwd", 	sk_passwd }, */
	{ "rm0",	sk_rm0},
	{ "startlog",	startlog},
	{ "rotlog",	rotlog},
  };

/*============================================================================*
 *			Execution of a command
 *============================================================================*/
static int comalog(time2, tmpfile, logo, prog)
/*++++++++++++++++
.PURPOSE  Log the parameters to a dedicated log file
.RETURNS  0 = success
.REMARKS
-----------------*/
  time_t time2[2]; 	/* IN: Begin / End of Transfer */
  char *tmpfile ;	/* IN: File to copy to log     */
  char *logo ;		/* IN: Program logo, e.g. "@/var/log/logfile"	*/
  char *prog ;		/* IN: Program  + Options + Parameters  */
{
  char buffer[4096], *b, *p, *e ;
  struct tm *T;
  int fh, fi, c=0, n ;

    /* Get first the name of log file */
    p = logo ;
    if (*p == '@') p++ ;
    for (b = buffer; isgraph(*p); b++) *b = *(p++) ;
    *b = 0 ;	/* buffer = name of log file */

    fh = open(buffer, O_WRONLY|O_APPEND|O_CREAT, 0644) ;
    if (fh < 0) {
	sk_log(LOG_ERR, "Can't open logfile: %s\n", buffer) ;
	return (-1) ;
    }

    /* Prepare the Record */
    T = gmtime(time2);		/* Edit the starting time */
    if (T->tm_year < 1000) T->tm_year += 1900;
    sprintf(buffer, "\n====%04d.%02d.%02d,%02d:%02d:%02dZ %lds ",
      T->tm_year, T->tm_mon+1, T->tm_mday, T->tm_hour, T->tm_min, T->tm_sec,
      time2[1] - time2[0]);
    b = buffer + strlen(buffer) ;
    if ((e = getenv("USER"))) strcpy(b, e) ;  b += strlen(b) ;
    *(b++) = '@' ;
    if ((e = getenv("HOST"))) strcpy(b, e) ;  b += strlen(b) ;
    if ((e = getenv("NODE"))) sprintf(b, "(%s)", e);  b += strlen(b) ;

    /* Add the parameters, and the exact program */
    sprintf(b, " %s\n", prog) ;

    fi = -1 ;
    if (tmpfile) {
	fi = open(tmpfile, 0, -1) ;
	if (fi < 0) {
	    sk_log(LOG_ERR, "Can't open tmp file: %s\n", tmpfile) ;
	    fi = -1 ;
	}
    }

    lockf(fh, 1, -1L) ; 	/* Lock File */
    write(fh, buffer, strlen(buffer)) ;
    if (fi >= 0) {
	while ((n = read(fi, buffer, sizeof(buffer))) > 0) {
	    write(fh, buffer, n) ;
	    c = buffer[n-1] ;
	}
	if (c != '\n') write(fi, "\n", 1) ;
	close(fi) ;
    }
    close(fh) ;

    return(0) ;
}

static int comlog(logo, exec, parm)
/*++++++++++++++++
.PURPOSE  Log a command
.RETURNS  0 = OK
-----------------*/
  char *logo;	/* IN: How to log: "-" = don't, "" = default */
  char *exec;	/* IN: What is exactly executed */
  char *parm;	/* IN: Supplementary parameter  */
{
    if (!logo)  return(0);		/* No logo   */

    switch(logo[0]) {
      case '-': return(0);	/* Don't log */
      case '@': return(0);	/* Don't log, there is a dedicated log file */
      case 0:			/* No logo, use EXEC */
    	sk_log(LOG_INFO, "EXEC %s%s\n", exec, parm);
    	break;
      case '*':			/* Don't log parameters */
	sk_log(LOG_INFO, "%s\n", logo);
    	break;
      default:			/* Use logo + parm.	*/
        while (isgraph(*exec)) exec++;
        while (isspace(*exec)) exec++;
	sk_log(LOG_INFO, "%s%s%s\n", logo, exec, parm);
    }
    return(0);
}

static int exec(coms, text, plug, opt)
/*++++++++++++++++
.PURPOSE  Execute a Command, shell of other, with complete tests & logging
.RETURNS  0 = OK
.REMARKS
-----------------*/
  COMS *coms;	/* IN: List of available commands 	*/
  char *text;	/* IN: Command to Execute 		*/
  int plug;	/* IN: Opened socket 			*/
  int opt;	/* IN: Option to allow direct calls to functions */
{
  char verbop;	/* Verbose option */
  char *parm, *parm1, x, *p, *red2;	/* Red2 = File 2 redirection */
  int  i, status;
  ACTION *pa;
  char errfile[40];		/* Used with command starting with ! */
  static char *line[] = {
	(char *)0, 	/* Command Name */
	(char *)0,	/* Command equivalence */
	"",		/* Who's allowed  */
	"-",		/* Don't log text */
	"",		/* Mnemonic parms */
	""		/* Help text	  */
  };
  time_t exec_time[2] ;

    verbop = ov;	/* Save current verbose option */
    exec_time[0] = exec_time[1] = 0 ;

	/* Translate the command if necessary */
    if ((opt == 0) || (isalnum(*text))) {
	parm = trcom(coms, text, sk_caller(plug));
    }
    else {	/* Do NOT translate. Do not log either! */
	parm = text;
	line[0] = line[1] = text;
	thecom = line;
    }
    if (!parm) goto Error_Occured;
    while(isspace(*parm)) parm++;
    pa   = (ACTION *)0;
    red2 = " 2>&1";	/* Redirection for Script */
    errfile[0] = 0;

    	/* The first char indicates local or shell function */
  ExamineCommand:
    switch(*parm) {
      case '+': /* Set VERBOSE Option */
	ov++;
	parm++;
	sk_servlog(1);
	goto ExamineCommand;
      case '*':	/* Local Function */
	/* For Complex Equivalences, split with ; */
	i = streloc(parm, ';') ;
	if ((x = parm[i])) {
	    parm1 = p = strdup(parm) ;
	    while (*p) {
		i = streloc(p, ';');
	        x = p[i] ; p[i] = 0 ;
	        if (exec(coms, p, plug, 1)) {
		    sk_log(LOG_ERR, "***%s error <%s> %s\n",
			parm, p, com_err);
		    sk_errsend(plug, "****Error in Complex Expression\n") ;
		}
		if (!x) break ;
		p[i] = x ;
		p += i+1 ;
		while (isspace(*p)) p++ ;
	    }
	    free(parm1) ;
	    break ;
	}
	if ((pa = find_action(parm, actions, ITEMS(actions)))) {
	    comlog(thecom[3], parm, "");
	    while(isgraph(*parm)) parm++ ;
	    while(isspace(*parm)) parm++ ;
	    for (p = parm + strlen(parm); --p >= parm; ) if (isgraph(*p)) break;
	    x = *++p; *p = 0;
	    exec_time[0] = time((time_t *)0) ;
	    if (setjmp(jmp_env) == 0)
	    	status = (*(pa->action))(parm, plug);
    	    else sk_log(LOG_WARNING, "Interrupted returned to %d\n", __LINE__),
	    	status = -1;
	    *p = x;
	    exec_time[1] = time((time_t *)0) ;
	    if (status < 0) pa = (ACTION *)0;
	}
	if (!pa) parm = (char *)0;
	break;
      case '!':		/* stderr redirected to a FILE */
	sprintf(errfile, " 2>/tmp/SV%d", getpid());
	red2 = errfile;
      case '-':		/* stderr is IGNORED */
	parm++;
	if (red2 != errfile) red2 = " 2>/dev/null";
			/* No Break  */
      default:		/*   Shell   Command */
	comlog(thecom[3], parm, red2);
	exec_time[0] = time((time_t *)0) ;
	if (shell(parm, red2, plug) != 0)
	    parm = (char *)0;	    /* Error */
	if (red2 == errfile)  	   /* Errors */
	    send_errors(plug, errfile+3);
	exec_time[1] = time((time_t *)0) ;
	break;
    }
  Error_Occured:
    if (!parm) {	/* Something wrong... */
      char buf[32], *b ;
      int len ;
        parm = strerror(com_errno);
        if (!parm) sprintf(parm=buf, "[unknown errno #%d]", com_errno) ;
	len = 4 + strlen(com_err) + strlen(parm) + strlen(text);
	b = malloc(len);
	sprintf(b, "%s%s: %s\n", com_err, text, parm);
        sk_log(LOG_ERR, "Error in exec(%s): %s", text, b);
	sk_errsend(plug, b);
	free(b);
	ov = verbop;
	return(-1);
    }
    if ((p = strchr(thecom[3], '[')))
	xferlog(exec_time, bytes_transfered,
	p, parm ? parm : "") ;
    if (*thecom[3] == '@') comalog(exec_time, sk_filename,
	thecom[3], parm ? parm : "") ;
    if (ov) sk_log(LOG_INFO, "Normal Return from exec(%s)\n", text);
    ov = verbop;
    return(0);
}

static int exec_init(coms)
/*++++++++++++++++
.PURPOSE  Execute the ".init" command
.RETURNS  0 = OK
-----------------*/
  COMS *coms;	/* IN: List of available commands */
{
  char *parm, *p, x;
  int i;
    parm = trcom(coms, ".init", "");
    if (!parm) return(0);	/* No initialisation... */
    while (isspace(*parm)) parm++;
    parm = strdup(parm); 	/* Address would collapse with exec */
    for (p = parm; *p; ) {
    	i = streloc(p, ';'); x = p[i]; p[i] = 0;
	/* Note: plug is set to -1, because plug is not required in init */
    	if (exec(coms, p, -1, 1)){   /* 1 = execute without check */
	    sk_log(LOG_ERR, "***.init error <%s> %s\n", p, com_err);
	    return(-1);
	}
    	p += i;
    	*(p++) = x;
    	if (!x) break;
	while(isspace(*p)) p++;
    }
    free(parm);
    return(0);
}

/*============================================================================*
 *			Cleanup the HOME directory
 *============================================================================*/
static int exec_cleanup(char *dirname)
/*++++++++++++++++
.PURPOSE  Clean up the directory containing temporary files
	(files starting by 'U' with unix time, or T/V)
.RETURNS  Number of removed files.
-----------------*/
{
  DIR *dirp; struct dirent *dire;
  char cwd[LEN_CWD];
  struct stat buf;
  time_t now, t;
  int ns, nf=0;
  char *p;

    cwd[0] = '?'; cwd[1] = 0;
    if (!getcwd(cwd, sizeof(cwd))) {
	sk_perror("exec_cleanup");
	sk_log(LOG_ERR, "exec_cleanup(%s): could not get cwd!\n", dirname);
	if ((p = getenv("HOME"))) {
	   sk_log(LOG_WARNING, "(assume cwd=$HOME=%s)\n", p);
	   strncmp(cwd, p, sizeof(cwd));
	}
	else return(0);
    }
    if (strcmp(dirname, cwd) == 0) {	/* No directory change */
	cwd[0] = 0;
	dirname = ".";
    }

    dirp = opendir(dirname);
    if (!dirp) {
	sk_perror("exec_cleanup");
	sk_log(LOG_WARNING, "Can't opendir '%s' (from '%s')\n",
		dirname, cwd);
	return(0);
    }
    now = time((time_t *)0);
    if (ov) sk_log(LOG_INFO, "....exec_cleanup(%s) at %lu\n",
	    dirname, now);

    if (chdir(dirname)) {
	sk_perror("exec_cleanup");
	sk_log(LOG_WARNING, "Can't chdir from '%s' to '%s'\n",
		cwd, dirname);
	closedir(dirp) ;
	return(0);
    }

    while ((dire = readdir(dirp))) {
	ns = dire->d_name[0];	/* First letter of file (V4.72) */
	if ((ns < 'T') || (ns > 'V')) continue;
	if (lstat(dire->d_name, &buf) < 0) {
	    sk_perror(dire->d_name);
	    sk_log(LOG_ERR, "exec_cleanup(%s)\n", dirname);
	    continue;
	}
	if (ov) sk_log(LOG_INFO, "    %s mod=%ld acc=%ld\n",
		dire->d_name, buf.st_mtime, buf.st_atime);
	if (!S_ISREG(buf.st_mode)) {
	    if (ov) sk_log(LOG_INFO, "    %s (not a file)\n", dire->d_name);
	    continue;
	}
	t = (time_t)atoi(dire->d_name+1);
	if (t>now) {	/* Not yet necessary to remove... */
	    if (ov) sk_log(LOG_INFO, "(too early, live for %lds)\n", t-now);
	    continue;
	}
	ns = now- buf.st_atime;	/* Seconds elapsed since last access */
	if (ns < (CHECKFILE_INTERVAL/2)) {
	    if (ov) sk_log(LOG_INFO, "(accessed since %ds)\n", ns);
	    continue;
	}
	ns = unlink(dire->d_name);
	nf++;
	if (ov) sk_log(LOG_INFO, " (unlink_status=%d)\n", ns);
    }
    closedir(dirp) ;

    /* Return in the original directory */
    if (cwd[0]) chdir(cwd);
    return(nf);
}

/*============================================================================*
 *			Signals
 *============================================================================*/

static void on_death(s)
/*++++++++++++++++
.PURPOSE  Deal with Zombie
.RETURNS  -
-----------------*/
  int s ;	/* IN: Unused signal */
{
  int st;
    /* sk_log(LOG_WARNING, "on_death signal %d\n", s); */
    signal(SIGCHLD, on_death);
    /* wait(&st) ;  -- THIS IS NOT ENOUGH -- several children may die
       more or less simultaneously
    */
    while (wait3(&st, WNOHANG, (struct rusage *) 0) > 0) ;
}

static void onIntr(s)
/*++++++++++++++++
.PURPOSE  Activated when INTR
.RETURNS  --
-----------------*/
  int s;	/* IN: Signal number */
{
    sk_log(LOG_WARNING, "onIntr  signal %d\n", s);
    signal(s, onIntr);
    interrupted = 1;
    if (toclose) close(toclose), toclose = 0;
    else longjmp(jmp_env, 1);
}

static void onHangup(s)
/*++++++++++++++++
.PURPOSE  Activated when Hangup
.RETURNS  --
.REMARKS  Used as a simple subroutine when s=0
-----------------*/
  int s;	/* IN: Signal number */
{
  COMS *coms;
  int st;

    sk_log(LOG_WARNING, "onHangup signal %d\n", s);
    if (s) {
        signal(s, onHangup);
        if (pid_root && (pid_root != getpid())) {
	    kill(pid_root, s);
	    return;
	}
    }

    /* Close/reopen the logfile */
    if (logfile && logname && s) {
	st = rename(logname, logname_0);
	if (!st) 			/* ERROR in rename */
	    sk_perror(logname_0);
	fclose(logfile);
	logfile = fopen(logname, "a");
	sk_setlog(logfile);
	/* Move the other files */
	if (st == 0) 			/* No error in rename */
	    rotlog();
    }

    /* Re-read the list of services */
    if (!ffile) return;
    coms = load_coms((COMS *)0, ffile);

    if (!coms) {
	sk_log(LOG_ERR, "command file disappeared (Hangup trial): %s\n", ffile);
	return;
    }
#if 0	/* No .init required -- it is executed by the CHILD */
    if (exec_init(coms)) {	/* FAIL ! */
	sk_log(LOG_ERR, "BAD command file (Hangup trial): %s\n", ffile);
	free_coms(coms);
	return;
    }
#endif
    free_coms(commands);
    commands = coms;	/* Now, list of commands reinstalled */
    sk_log(LOG_WARNING, "Successfully reloaded commands '%s'\n", ffile);
    if (s) periodic_check(0);	/* Do also the cleanup */
}

static void periodic_check(s)
/*++++++++++++++++
.PURPOSE  Verify the config file changed
.RETURNS  --
-----------------*/
  int s;	/* IN: Signal number */
{
  struct stat buf;
  int day;

    if (ov) sk_log(LOG_INFO, "alarm raised...\n");
    if (s) {
        signal(s, periodic_check);
        alarm(CHECKFILE_INTERVAL);
    }
#if 0		/* Old version: external code */
    if (cleanup_command) {
      FILE *pop; char buf[256];
	sk_log(LOG_DEBUG, "Executing: %s\n", ov ? cleanup_command : "Cleanup");
	pop = popen(cleanup_command, "r");
	if (pop) {
	    while(fgets(buf, sizeof(buf), pop))
		sk_log(LOG_DEBUG, "%s", buf);
	     fclose(pop);
	}
	else sk_perror(cleanup_command);
        /* system(cleanup_command); */
    }
#else
    exec_cleanup(getenv("HOME"));
#endif

    /* Rotate the logs ?
     * sighup is used in onHangup as 
     * 0 = open/close log
     * sighup = rotate logs, reload conf. file, etc
     **/
    day = time((time_t *)0)/86400.;
    if (logname && (day != jour)) {
	jour = day;
	if (s) onHangup(SIGHUP);
	return;
    }
    if (!ffile) return;
    if (stat(ffile, &buf)) 		/* Error */
	return;
    if (buf.st_mtime == ffile_time)	/* No change */
	return;
    if (s) onHangup(0);			/* Reload Config File */
}

/*============================================================================*
 *			Main Program
 *============================================================================*/
int main( argc, argv)
/*++++++++++++++++
.PURPOSE  Server executing sh commands
.RETURNS  to OS
.REMARKS  Usage:
-----------------*/
  int argc; char **argv;
{
  static char *home_env = "HOME=/tmp";
  static FILE *pidfile;
  static int cumulerrors = 0;
  static int cumulmax = 10;		/* Stop after that amount .. */
  int plug, sock, gid, uid, i;		/* fd for the working socket */
  char buf[BUFSIZ];
  char *parm, *p;
  static char *service = (char *)0;
  static char od, olaunch;	/* Options */
  static int pendent;

	/* SK_bsk may contain the physical size of socket blocks */

    if ((p = getenv("SK_bsk"))) {
	i = atoi(p);
	while(isdigit(*p)) p++;
	if (tolower(*p) == 'K') i *= 2;
	sk_setb(i);
    }
    putenv(home_env);

	/* Check the Arguments */

    logfile = (FILE *)0;
    gid = uid = 0;
#   ifdef DEBUG
    /* logfile = fopen("/tmp/xbd", "w"); */
    logfile = stderr;
    fprintf(logfile, "argc=%d\n", argc);
#   endif
    while (-- argc > 0) {
	parm = *++argv;
#   ifdef DEBUG
	printf("....Now argc=%d, argv=%s, parm=%s\n", argc, *argv, parm);
#   endif
	if (logfile) fprintf(logfile, "Parm=%s\n", parm), fflush(logfile);
	if (*parm == '-') while (*++parm) switch (*parm) {
	  case 'A':	/* /etc/passwd file   */
		afile = "";
		continue;
	  case 'a':	/* Authorisation File */
		--argc;
		afile = *++argv;
		if (strcmp(afile, ".") == 0) afile = "";	/* Default */
		parm = "\0\0";		/* To force continue outer loop */
		continue;
	  case 'c':	/* Pendent connections */
		--argc; ++argv;
		pendent = atoi(*argv);
		parm = "\0\0";		/* To force continue outer loop */
		continue;
	  case 'f':	/* Table of Functions */
		--argc;
		ffile = *++argv;
		if ((commands = load_coms(commands, ffile)))
		    exec_init(commands);
		parm = "\0\0";		/* To force continue outer loop */
		continue;
	  case 'b':	/* Set Socket Size */
		--argc; ++argv;
		sk_setb(atoi(*argv));
		parm = "\0\0";		/* To force continue outer loop */
		continue;
	  case 'l':	/* Look for log */
		if (strcmp(parm, "launch") == 0) {
		    olaunch = 1;
		    if (!pendent) pendent = 5;	/* Default */
			parm = "\0\0";	/* To force continue outer loop */
		    continue;
		}
		--argc; ++argv;
		if (logfile) fclose(logfile);
		logfile = fopen(logname = *argv, "a");
		sk_setlog(logfile);
		logname_0 = malloc(strlen(logname)+3);
		strcpy(logname_0, logname); strcat(logname_0, "..");
		parm = "\0\0";		/* To force continue outer loop */
		continue;
	  case 'p':	/* pidfile */
		--argc; ++argv;
		pidfile = fopen(*argv, "w");
		if (!pidfile) {
		    fprintf(stderr, "****");
		    perror(*argv);
		    exit(1);
		}
		continue;
	  case 'D':	/* HOME  */
		--argc; ++argv;
		if (chdir(*argv)) {
		    fprintf(stderr, "****cd ");
		    perror(*argv);
		    exit(1);
		}
                home_env = malloc(strlen(*argv)+6);
                strcpy(home_env, "HOME=");
                strcat(home_env, *argv);
                putenv(home_env);
		continue;
	  case 'd':	/* Inetd */
	  case 'i':	/* Inetd */
		od = 1; continue;
	  case 't':	/* Time-out */
		--argc; ++argv;
		p = *argv;
		duration_period = atoi(p);
		while(isdigit(*p)) p++;
                if (isalpha(*p)) {
		    for (i=0; duration_suffix[i]; i += 2) {
	                if (*p == duration_suffix[i]) { i++; break; }
	            }
	            if (duration_suffix[i])
			duration_period *= duration_suffix[i];
                }
		continue;
	  case 'v':	/* Verbose */
		if (strncmp(parm, "ver", 3) == 0) {
		    printf("aserver (CDS) Version %s\n", VERSION);
		    exit(0);
		}
		ov++; continue;
	  case 'x':	/* Additional command */
		--argc; ++argv;
		p = *argv;
		if (p && (*p == '+')) p = more_conf;
		commands = append_coms(commands, p);
		exec_init(commands);
		continue;
	  case 'h':	/* Help */
		printf(usage, pgm);
		printf("%s", help);
		exit(0);
	  case 'u':	/* uid */
		++parm;			/* No Break */
	  default:
		if (isdigit(*parm)) {	/* UID */
		    uid = atoi(parm);
		    while(isdigit(*parm)) parm++;
		    if (ispunct(parm[0]) && isdigit(parm[1])) {	/* GID */
		    	++parm; gid = atoi(parm);
		    	while(isdigit(*parm)) parm++;
		    }
		    --parm;
		    continue;
		}
		error("Bad option: ", parm);
	} else {
		if (service) error("Bad option: ", parm);
		else service = parm;
	}
    }
    if (!service) error("Missing service", "");
    if (!commands) commands = append_coms(commands, "");

#   ifdef DEBUG
    if(logfile) {
	fprintf(logfile, "Terminated Options\n"),
	fprintf(logfile, "Service is %s", service);
	fflush(logfile);
    }
#   endif


	/* Set up the Options for sk_start */
    p = buf;
    for (i=0; i<ov; i++) *(p++) = 'v';
    if (od) {
	*(p++) = 'd';
	if (!logfile) *(p++) = 'l';	/* Use Syslog */
	olaunch = 0;	/* Incompatibility */
    }
    *p = 0;

	/* Start the Server */
    if (logfile)
	fprintf(logfile, "... Ready for sk_start ...\n"), fflush(logfile);
    sock = sk_astart(service, buf, afile);
    if (logfile)
	fprintf(logfile, "... sk_start returned %d\n", sock),
	fflush(logfile);
    if (sock < 0) exit(1);

    /* Change of uid can be done here if no password checking */
    if ((!afile) && uid) {
	if (gid) setgid(gid);
	if (setuid(uid) < 0) {
	    sk_perror("ABORT setuid");
	    exit(1);
	}
	sk_log(LOG_DEBUG, "SUID to %d\n", uid);
	uid = 0;
    }

    if (olaunch) {
	if (fork()) _exit(0);		/* End of father process */
	signal(SIGCHLD, on_death);
	signal(SIGHUP, onHangup);
	signal(SIGALRM, periodic_check);
	signal(SIGUSR1, periodic_check);
	alarm(CHECKFILE_INTERVAL);
    	fprintf(stderr, "----Process launched: %d\n", getpid());
	cumulmax = 999999 ; 		/* Accept 1 million errs */
    }
    pid_root = getpid();
    if (pidfile) {
	fprintf(pidfile, "%d\n", pid_root);
	fclose(pidfile);
    }
    jour = time((time_t *)0)/86400;	/* Day (change of day rotates logs) */

    while(1) {			/* Loop in Waiting for Connections */
#ifdef DEBUG
	printf("----sk_listen(%d, %d)\n", sock, pendent);
#endif
    	plug = sk_listen(sock, pendent);
#ifdef DEBUG
    	if (logfile)
	    fprintf(logfile, "... sk_listen returned %d\n", plug),
	    fflush(logfile);
#endif
    	if (plug < 0) {
	    ++cumulerrors;
	    sk_log(LOG_ERR,
	      "Bad return (#%d) from sk_listen -- Give up!\n",
	      cumulerrors);
	    if (cumulerrors >= cumulmax) {
		sk_log(LOG_ERR,
		  "Too many (%d) sk_listen errors -- TERTMINATE ****",
		  cumulerrors);
		exit(1);
	    }
	    else continue;
	}
	cumulerrors = 0;	/* Any error state is over.. */
	if (olaunch) {
	    if (fork()) {	/* ================== Father */
		close(plug);
		continue;
	    }
	    close(sock);	/* ================== child  */
	    signal(SIGALRM, duration_expired);
	    alarm(TIME_MAX);	/* -- Changed by sk_duration */
	}
    	p = thecaller = sk_caller(plug);
    	sk_log(LOG_INFO, "CONNECT from %s\n", p);
    	setuser(p);
    	if (uid) {
	    if (gid) setgid(gid);
	    if (setuid(uid) < 0) {
		sk_perror("ABORT setuid");
		exit(1);
	    }
	    sk_log(LOG_DEBUG, "SUID to %d\n", uid);
    	}

	/* Define umask for file creation */
    	umask(022);

	/* Execute the Initialisation */
    	if (commands) {
	    i = exec_init(commands);
	    if (i) exit(1);	/* FAIL */
    	}

    	/* read lines until the stream closes */
	if (!od) {
	    sk_urgent(plug);	/* Install URGENT */
	    sk_log(LOG_DEBUG,"URGENT installed\n");
	    signal(SIGINT, onIntr);	/* Install INTERRUPT handler */
	}
    	while (1) {
	    if (setjmp(jmp_env) == 0)  {
  	    	i = sk_gets( plug, buf, sizeof(buf)-4);
	    	if (i <= 0) break;
	    	buf[i] = buf[i-1] = 0;
	    	if(ov) sk_log(LOG_INFO,  "Read <%s>\n", buf );
	    	interrupted = 0;
	    	if (exec(commands, buf, plug, 0) < 0) {	/* BAD RESULT */
        	    sk_log(LOG_ERR, "ERROR STOP\n");
		    sk_errsend(plug, "****Stopped execution\n");
		    exit(1);
		}
	    	/* if (exec(commands, buf, plug, 0))
	    	    buf[i-1] = '\n', sk_errsend(plug, buf); */
	    	if (eof) break;
	    }
    	    else sk_log(LOG_WARNING, "Interrupted returned to %d\n", __LINE__),
	    	sk_log(LOG_DEBUG, "... Sending ^D (\04)\n");
	    sk_put(plug, "\04");	/* END Indicator */
    	}
    	sk_log(LOG_INFO, "QUIT\n");
    	sk_log(-1);			/* CLOSE log */
    	exit(0);
    }
}
