/*++++++++++++++
.IDENTIFICATION skserver.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    Internet
.KEYWORDS       Client/Server, TCP/IP
.VERSION  1.0   07-Aug-1992
.VERSION  1.1   20-Aug-1992: Added sk_fromclient, sk_toclient
.VERSION  1.2   02-Oct-1992: Corrected bug for DEC Machines 
			in sk_start
.VERSION  2.0   10-May-1993: Added ^C (error transmitted to Client).
.VERSION  2.1   28-May-1993: Copy authorisation file into new piece of memory
			Server returns pid.sk to Client as Ackowledgment
.VERSION  2.2   13-Jul-1993: Don't send \n 's in error messages
.VERSION  2.3   29-Jul-1993: listen anly called once
.VERSION  2.4   25-Sep-1993: possibility of logging into /tmp/fromclient
.VERSION  2.5   20-Dec-1993: Tested DEC-ALPHA
.VERSION  2.6   28-Jan-1994: Use /etc/passwd
.VERSION  2.7   27-Sep-1994: Check if client sends back correctly Ack.
.VERSION  2.8   03-Nov-1994: Bufferisation
.VERSION  2.9   16-Jan-1995: BASIC option. No sk_setlog
.VERSION  2.91  10-Feb-1995: better logging in case of Writing Error
.VERSION  2.92  10-Mar-1995: (still) better logging in case of Writing Error
.VERSION  2.93  10-Oct-1995: Add shadow access to passwd() - MW
.VERSION  2.94  07-Dec-1995: use signal blocking in sk_listen() - MW
.VERSION  3.0   12-Dec-1995: split sk_listen() in sk_accept() & sk_ack()
.VERSION  3.1   30-Jun-1998: A BUG discovered and corrected in umatch !!!
			sk_accept() should not keep any io with the client 
			sk_ack() should be called from the child after the fork.
.VERSION  3.1   30-Jun-1998: A BUG discovered and corrected in umatch !!!
.VERSION  3.2   04-Oct-2003: Use stdarg
.VERSION  3.3   04-Nov-2003: sk_maxbytes to limit the size of transfers
.VERSION  3.4   04-Mar-2004: Write out the port mapping
.VERSION  3.41  14-Feb-2007: Changed value NSOCKS 64 --> 256
.COMMENTS       Routines for a TCP/IP Server 
	Adapted from Blair P. Houghton
	Routines for the Server part include:
	sk_start (service, mode)      	Start a Service
	sk_astart(afile, service, mode) Start a Service with Authorisation
	sk_listen(sk, queue_slots)	Wait for incoming calls

	Other routines:
	sk_log(level, fmt, arguments)	Log Server Messages
	sk_caller(plug)			Tell who is connected

---------------*/


/* The VSYSLOG function exists on Solaris and Sun */
#ifndef VSYSLOG
#ifdef SOLARIS
#define VSYSLOG		1
#else
#define VSYSLOG		0
#endif
#endif

/* Definitions concerning the Shadow file */
#ifdef SOLARIS
#include <shadow.h>
#endif

#ifndef NSOCKS
#define NSOCKS 256
#endif
#include <sk.h>			/* Local definitions		*/
#include <skserv.h>		/* Local definitions (server)	*/
#include <pwd.h>		/* /etc/passwd facilities	*/
#include <time.h>		/* time() function		*/
#include <errno.h>		/* system error messages	*/
#include <stdio.h>
#include <fcntl.h>		/* file mnemonics		*/
#include <netdb.h>		/* gethostby... definitions	*/
#include <ctype.h>		/* isdigit, etc...		*/
#include <string.h>
#include <signal.h>
#include <stdlib.h>		/* malloc, etc...		*/
#include <unistd.h>		/* getpid, etc...		*/
#include <syslog.h>		/* System Log Facility		*/
#include <sys/file.h>		/* For File Locking		*/
#include <sys/types.h>		/* Standard data types		*/
#include <sys/stat.h>		/* For tapes, etc		*/
#ifndef BASIC
#include <sys/mtio.h>		/* For tapes, etc		*/
#endif
#include <sys/ioctl.h>		/* SIO definitions              */
#include <sys/socket.h>		/* Socket Definitions		*/
#include <netinet/in.h>		/* Internet sockets definitions	*/
#include <arpa/inet.h>		/* inet_... functions		*/
#ifdef __osf__		/* DEC-ALPHA */
# include <arpa/inet.h>
#endif
#ifdef SOLARIS
# include <arpa/inet.h>
# include <stdarg.h>		/* Variable number of Arguments	*/
/*# include <stdarg.h>		-- Variable number of Arguments	*/
#else
# include <stdarg.h>		/* Variable number of Arguments	*/
#endif
#include <netinet/tcp.h>	/* Internet sockets definitions	*/
#ifndef SIOCATMARK		/* For Solaris ...		*/
#include <sys/sockio.h>		/* Socket Definitions		*/
#endif

#ifndef LOG_DAEMON		/* e.g. Dec has not LOG_DAEMON in syslog.*/
#define LOG_DAEMON 0
#endif
#define mexp(b, l)	(b ? realloc(b,l) : malloc(l))	/* Reallocation */

/* char *strdup(), *inet_ntoa(); */
typedef void (*FCT)();		/* Pointer to Function */
typedef int  (*iFCT)();		/* Pointer to Function */

static int theplug, thefile;
#ifndef BASIC
static struct mtget mt_stat;	/* MagTape Status */
static struct mtop  mtop;	/* MagTape Status */
#endif

#define OLAUNCH_vv	0x80	/* Very Verbose   */
#define OLAUNCH_v	0x40	/* Verbose option */
#define OLAUNCH_l	0x20	/*   Log   option */
#define OLAUNCH_d	0x10	/*  Daemon option */
#define OLAUNCHED	0x01	/*  Just launched */
static char ocalls[NSOCKS];	/* Mode options */
static char *calls[NSOCKS];
static char *ncalls[NSOCKS];	/* for caller address */
static char *routine = "";
static short routine_no = 0;	/* Service Number */
static FILE *deblog;		/* DEBUG log */
static char *afile;		/* Authorisation File; as
	either a true file name (alternate passwd file) 
	    or NULL (no passwd checking)
	    or ""   (/etc/passwd file)
	*/

/* To solve the problem of large files, once the number MAXBYTES
   is reached, the number of blocks is stored in blocks .
   This number can be retrieved by sk_blocks
*/
static int4 blocks ;		/* Count blocks if < 2GBytes */
#define MAXBYTES   0x7fff0000	/* Getting too close to 2GBytes */

static int maxbytes = 0;	/* Set by sk_maxbytes */

char *sk_caller();
FCT sk_errfct();
iFCT authentification;
FILE *sk_getlog();

static int lockfile(file, opt)
/*++++++++++++++++
.PURPOSE  Lock a file
.RETURNS  0 (success) / -1 (failure)
-----------------*/
  FILE *file;		/* IN: File to lock / unlock */
  int opt;		/* IN: Option 1=Lock 0=Unlock	*/
{
#ifndef BASIC		/* Lock is normally not required */ 
    int trylocks = 10;
    while (trylocks > 0) {	/* Verify the Lockfile */
	if (flock(fileno(file), opt ? LOCK_EX|LOCK_NB : LOCK_UN) == 0) 
		return(0);
	sleep(1);
    }
#else
    return(-1);
#endif
}

static int strloc(text, c)
/*++++++++++++++++
.PURPOSE  Locate specified character
.RETURNS  Index of located char
-----------------*/
  char *text;     /* IN: String to interpret */
  int  c;         /* IN: Char to look for */
{
  char *s;
    for ( s = text; *s; s++)      if (*s == c)    break;
    return(s-text);
}

static int tmatch(template, name)
/*++++++++++++++++
.PURPOSE  Check name against template
.RETURNS  1 (Match) / 0 (mismatch)
.REMARKS  Only * ? recognized as wild chars
-----------------*/
  char *template;	/* IN: Name template with * ? */
  char *name;		/* IN: Name to check */
{
  char *t, *f;

    for (t = template, f = name; *t; t++) switch(*t) {
      case '?':			/* Matches one character */
	if (*f) f++;
	else return(0);
	continue;
      case '*':			/* Match 0 or n characters */
	if (! *++t) return(1);	/* The * is the last char of template */
	for (; *f; f++) {	/* Look for first matching char */
	    if (*f != *t) continue;
	    if (tmatch(t,f))	/* Recursive call */
		return(1);
	}
	return(0);
      default:
	if (*t != *f) return(0);
	f++;
	continue;
    }
    return((!*t) && (!*f));	/* Returns 1 if *t AND *f are NUL */
}

/*============================================================================*/
int sk_log(int lev, ...)
/*++++++++++++++++
.PURPOSE  Write errors, either on syslog, or on another file
.RETURNS  The level (first argument)
.REMARKS  First Argument = Level / Second Argument = fmt / Other = arguments
	The level is defined in syslog.h from 0 (emergency) to ? (debug)
	A value of -1 for the level asks to CLOSE the log.
-----------------*/
{
  va_list(args);		/* Variable List of Arguments */
  FILE *stdlog;
  char *fmt;
  static char alev[5] = "++++";
  int trylocks;

    va_start(args, lev);	/* First Argument */
    /* lev = va_arg(args, int); */
    fmt = va_arg(args, char *);

    if (lev <= LOG_ERR) 		alev[0] = '*';
    else if (lev <= LOG_WARNING) 	alev[0] = '+';
    else if (lev == LOG_NOTICE)  	alev[0] = '=';
    else if (lev == LOG_INFO)  		alev[0] = '-';
    else if (lev == LOG_DEBUG)  	alev[0] = '.';
    else 				alev[0] = '?';
    alev[1] = alev[2] = alev[3] = alev[0]; 

    if (deblog) {
	fprintf(deblog, "sk_log: lev=%d, fmt=%s\n", lev, fmt);
	fflush(deblog);
    }
    if ((stdlog = sk_getlog())) { 	/* Log in a file or stderr */
      	if (lev < 0)	{ if (stdlog != stderr) fclose(stdlog); }
      	else {
	    time_t tim; char *ct, *ctime();
	    time(&tim); ct = ctime(&tim); ct[20] = 0; 
	    trylocks = lockfile(stdlog, 1);
	    fprintf(stdlog, "%s%s%s[%d]: ", ct+4, alev, routine, getpid());
	    vfprintf(stdlog, fmt, args), fflush(stdlog);
	    if(trylocks == 0) lockfile(stdlog, 0);
   	}
    }
    else {
	if (lev < 0)	closelog();
#if 1		/* V3.41: Problems with syslog */
	if (isatty(2)) {
	    fprintf(stderr, "%s%s[%d]: ", alev, routine, getpid());
	    vfprintf(stderr, fmt, args) ;
	}
#else		/* Pre-V3.41 */
#if VSYSLOG  /* Use the function */
	else { 
	    vsyslog(lev, fmt, args); 
	    if (isatty(2)) {
		fprintf(stderr, "%s%s[%d]: ", alev, routine, getpid());
	    	vfprintf(stderr, fmt, args) ;
	    }
	}
#else		/* Supply an alternative */
    	else { char *arg[5]; int i; 
	    if (isatty(2)) {
		fprintf(stderr, "%s%s[%d]: ", alev, routine, getpid());
	    	vfprintf(stderr, fmt, args) ;
	    }
	    for (i=0; i < 5; i++) arg[i] = va_arg(args, char *);
    	    if (deblog) {
		fprintf(deblog, "(Emulation of vsyslog)");
		fprintf(deblog, fmt, arg[0], arg[1], arg[2], arg[3], arg[4]);
		fflush(deblog);
            }
	    syslog(lev, fmt, arg[0], arg[1], arg[2], arg[3], arg[4]);
#endif
	}
#endif		/* VSYSLOG */
    }
    va_end(args);
    return(lev);
}

void sk_perror(s) 
/*++++++++++++++++
.PURPOSE  Interface to perror
.RETURNS  -
-----------------*/
  char *s; /* IN: Message */
{
  char buf[32], *msg ;
    msg = strerror(errno);
    if (!msg) sprintf(msg=buf, "[unknown errno #%d]", errno) ;
    if (deblog) fprintf(deblog, "%s: %s\n", s, msg),
	fflush(deblog);
    sk_log(LOG_ERR, "%s: %s\n", s, msg) ;
}

/*============================================================================*
 *		Signal Processing
 *============================================================================*/
int usig(plug)
/*++++++++++++++++
.PURPOSE  Find out what is the signal passed by Client
.RETURNS  The Signal (0 = no signal)
.REMARKS  The other incoming messages are thrown away
-----------------*/
  int plug;	/* IN: opened socket */
{
  char buf[2028]; int stat;

#ifdef DEBUG
	printf("... Entered  usig plug#%d\n", plug);
#endif
    while(1) {
	if (ioctl(plug, SIOCATMARK, &stat) < 0) {
	    sk_perror("ioctl(SIOCATMARK)");
	    return(-1);
	}
	if (stat == 1)	break;		/* Passwd OOB char found */
	recv(plug, buf, sizeof(buf), 0);
#ifdef DEBUG
	printf("... read <%s>\n", buf);
#endif
    }
    if (recv(plug, buf, 1, MSG_OOB) < 0) {
	sk_perror("recv(SIGNAL)");
	return(-1);
    }
    return(buf[0]);
}

static void on_urgent(sig)
/*++++++++++++++++
.PURPOSE  URGENT routine (signal sent by Client)
.RETURNS  --
.REMARKS  Executes the appropriate routine. 
-----------------*/
  int sig;	/* IN: Normally SIGURG */
{
  int val;
    signal(sig, on_urgent);
    val = usig(theplug);
    sk_log(LOG_WARNING, "URGENT signal (%d) received on plug#%d\n", 
	val, theplug);
    if (val > 0) {
    	sk_log(LOG_WARNING, " ====> kill(%d, %d) \n", getpid(), val);
	kill(getpid(), val);
    }
}

void sk_urgent(plug)
/*++++++++++++++++
.PURPOSE  Install a Urgent trapping
.RETURNS  --
-----------------*/
  int plug;	/* IN: Soket number */
{
    theplug = plug;
    signal(SIGURG, on_urgent);
    fcntl(plug, F_SETOWN, getpid());
}

/*============================================================================*
 *		Check the Userid / Password
 *============================================================================*/

static int passwd(user, pswd)
/*++++++++++++++++
.PURPOSE  Routine to check User / Password
.RETURNS  -1 = Bad / 0 = OK
.REMARKS Use the /etc/passwd file UserName:Password:uid:gid: gecos : 
-----------------*/
  char *user;	/* IN: RemoteName@Internet(host):Username */
  char *pswd;	/* IN: Associated Password */
{
#ifdef SOLARIS
  struct spwd *spw;	/* Shadow */
#endif
  struct passwd *pw;
  char *p, *u, *crypt();
  int stat;

    if (!afile) {
	sk_log(LOG_WARNING, "Authorize File missing for Intruder %s\n", user);
	return(0);
    }

    if (*afile) {	/* Non-standard passwd file */
#ifdef SETPW		/* The function setpwfile exists */
	getpwuid(0);	/* Must read at least one entry before change */
	if (! access(afile, 0)) {
	    sk_perror(afile);
	    afile = "";			/* Replace by standard passwd file */
	}
	else setpwfile(afile);  	/* Change passwd file */
#else
    sk_log(LOG_ERR, "%s access file not allowed, use /etc/passwd\n");
#endif
    }

    u = user + strloc(user, ':');
    if (*u) u++;
    else    u = "guest";	/* Default Name  */
    
    pw = getpwnam(u);
    if (! pw) {
    	sk_log(LOG_INFO, "NO USER %s in passwd file %s\n", u, afile);
    	return(-1);
    }

    	/* Check Password */
#ifdef SOLARIS
    spw = getspnam(u);
    p = spw->sp_pwdp;
#else
    p = pw->pw_passwd;
#endif
    stat = strcmp(crypt(pswd, p), p);
    if (stat) {
    	sk_log(LOG_INFO, "wrong PASSWORD for user %s in passwd file %s\n", 
	    user, afile);
    	return(-1);
    }
    	/* Check Remote Access */
    p = pw->pw_gecos;
    if (sk_umatch(user, p))	return(0);
    sk_log(LOG_INFO, "No access for %s in pwfile %s\n", user, afile);
    sk_errsend(theplug, "****Restricted Access (Sorry)");
    return(-1);
}

/*============================================================================*/
int sk_umatch(str, tem)
/*++++++++++++++++
.PURPOSE  Compare a string s (User@Internet(Machine)) with a template t
.RETURNS  0 if mismatch
.REMARKS  Template may include wild characters * or ?. Possible locations
		are separated by vlanks or commas.
-----------------*/
  char *str;	/* IN: String to check */
  char *tem;	/* IN: Template, * can match anything BUT dots */
{
  char x, *t, *s, *sm;
  int trials, i, match;
#ifdef DEBUG
    fprintf(stderr, "....umatch <%s> with template <%s>\n", str, tem);
#endif
	/* Take a copy of the two strings User@Internet  and  User@Machine */
    s = malloc(2*strlen(str)+2);
    strcpy(s, str); i = strloc(s,'('); s[i] = 0;
    sm = s + i +1; strcpy(sm, str);
    t = sm + strloc(sm, '@');
    if (*t && str[i]) strcpy(t+1, str+i+1);
    sm[strloc(sm, ')')] = 0;

    for (match = trials = 0; *tem && (!match); tem++)  {
	while (isspace(*tem)) tem++;
						/* Locate end of word */
    	for (t = tem; isgraph(*t) && *t != ',' ; t++) ;  
	x = *t; *t = 0;
	i = strloc(tem, '@');
	if (tem[i]) {				/* Word is user@host  */
	    trials++;
	    match = tmatch(tem, isalpha(tem[i+1]) ? sm : s);
	}
	if (!x) break ;
	*t = x;
	tem = t;
    }
    free(s);

    return(match ? match : trials == 0);	/* If @ never appears */
}

/*============================================================================*/


iFCT  sk_setauth(routine)
/*++++++++++++++++
.PURPOSE  Define the Authentification routine
.RETURNS  Previous routine
.REMARKS  The authentification routine must return -1 for Invalid Code
-----------------*/
  iFCT routine;	/* IN: Function (user@internet/serv:remote-user, pw) */
{
  iFCT old;

  old = authentification; 
  authentification = routine;
  return(old);
}

/*============================================================================*/

int sk_errsend(plug, msg)
/*++++++++++++++++
.PURPOSE  Send an error message to the Client
.RETURNS  0 / -1
-----------------*/
  int plug;	/* IN: Number returned by sk_listen */
  char *msg;	/* IN: Message to send		*/
{
  int stat; char buf[12];
    if (sk_write(plug, "\03", 1) != 1) return(-1);
    stat = sk_read(plug, buf, sizeof(buf));	/* Read Ack */
    if (stat < 0) return(stat);
    	/* Send Counted Buffer to Client */
    stat = strlen(msg);
    if (sk_putl(plug, stat) < 0) { stat = -1; return(-1); }
    if (stat == 0) return(0); 	/* END indicator */
    if (sk_write(plug, msg, stat) != stat) return(-1);
    return (0);
}

int sk_perrsend(plug, s)
/*++++++++++++++++
.PURPOSE  Send a system error message (perror) to the Client
.RETURNS  0 / -1
-----------------*/
  int plug;	/* IN: Number returned by sk_listen */
  char *s; 	/* IN: Message */
{
  char buf[32], *b, *msg ;
  int stat, lb ;
    msg = strerror(errno);
    if (!msg) sprintf(msg=buf, "[unknown errno #%d]", errno) ;
    lb = strlen(s) + 4 + strlen(msg);
    b = malloc(lb); sprintf(b, "%s: %s\n", s, msg);
    stat = sk_errsend(plug, b); 
    free(b);
    return (stat);
}

/*============================================================================*/

char *sk_caller(plug)
/*++++++++++++++++
.PURPOSE  Get caller coordinates as  User@Internet(Host)/Service_no:r-Who
.RETURNS  Pointer to relevant caller coordinates / NULL if error
.REMARKS  
-----------------*/
  int plug;	/* IN: Number returned by sk_listen */
{
    if ( (plug < 0) || (plug >= NSOCKS) ) return((char *)0);
    return (calls[plug]);
}

/*============================================================================*/

int sk_start(service, mode)
/*++++++++++++++++
.PURPOSE  Start a TCP Server, to a specified port or service, either
	directly or via inetd (mode includes 'd')
.RETURNS  The socket of file descriptor number / -1 if error
.REMARKS  Mode is a string made of the folowing chars:
	. l  means that syslog is used for logging  
	     (default: stderr if sk_setlog was not called)
	. v  verbose (debugging) mode
	. d  start as a Daemon (started from inetd)
	In "inetd" mode, upon successfull startup, sk_start returns a 
		number immediately usable by sk_read/sk_write.
	In "standalone" mode, sk_start returns a number to be used by
		sk_listen.
-----------------*/
  char *service;	/* IN: Service or Port Number */
  char *mode;	/* IN: Mode: d=Daemon,  v=Verbose l=SysLogging */
{
  int sock;			/* fd for the listening socket */
  struct sockaddr_in sockaddr;	/* sytem's location of the socket */
  struct servent *the_service;
  int sockaddr_in_length = sizeof(struct sockaddr_in);
  int iport=0, oport;
  char msg[64];
  char *p, om;


    routine = service; 	/* Set up the Routine name for Logging */

	/* Extract the Mode Options */
    om = 0;
    for (p = mode; *p; p++) switch(*p) {
	case 'l': om |= OLAUNCH_l; continue;
	case 'v': 
		if (om&OLAUNCH_v) om |= OLAUNCH_vv;
		else om |= OLAUNCH_v; 
		continue;
	case 'd': om |= OLAUNCH_d; continue;
	default: sk_log(LOG_ERR, "Bad mode <%s> in sk_start service <%s>\n", 
		mode, service);
	    return(-1);
    }

	/* Check the logging facilities */
    if (om&OLAUNCH_l) {
	sk_setlog((FILE *)0);
	openlog(service, LOG_PID, LOG_DAEMON);
#ifdef DEBUG
	sk_servlog(1);
	fprintf(deblog, "=================== Starting Daemon =============\n");
	sk_perror("OpenLog");
#endif
    }

    if(om&OLAUNCH_v) 
	sk_log(LOG_DEBUG, "Start service %s, mode %s\n", service, mode); 
    
    if (om&OLAUNCH_d) sock = 0;	/* Daemon Mode (started by inetd) */
    else	{ 		/* open a net socket, using stream mode, 
	     		   	with protocol irrelevant ( == 0 )	   */
	sockaddr.sin_family = AF_INET;
	sockaddr.sin_addr.s_addr = INADDR_ANY;	/* not choosy about who calls */
	if (isalpha(*service)) {	/* V3.41: Verify service exists */
	    the_service = getservbyname(service, (char *)0);
	    if (!the_service) {
	        sk_perror(service);
		return(-1);
	    }
	    sockaddr.sin_port = the_service->s_port;
	    iport = ntohs(sockaddr.sin_port);
	}
	else {
	    iport = atoi(service);
	    sockaddr.sin_port = htonl(atoi(service));
	}
#if 0		/* Old code */
	iport = atoi(service);
	sockaddr.sin_port = isalpha(*service) ?
	    (getservbyname(service, (char *)0))->s_port :
	    htonl(atoi(service));
#endif
    	if ( (sock = socket( AF_INET, SOCK_STREAM, 0 )) < 0 ) {
	    sk_perror("Can't assign socket");
	    return(-1);
        }
	 if (sock >= NSOCKS) {
	     sk_log(LOG_ERR, "Ear socket %d too large (>= %d)\n", sock, NSOCKS);
	     return(-1);
	}

    	if ( bind(sock, (struct sockaddr *) &sockaddr, sizeof sockaddr ) < 0 ) {
	    sk_perror("Can't bind socket");
	    return(-1);
    	}
    }

   		/* Get port number */
    if ( getsockname( sock, (struct sockaddr *) &sockaddr,
       (int *)&sockaddr_in_length ) < 0 ) {
	sk_perror("Can't getsockname");
	return(-1);
    }
    routine_no = oport = ntohs(sockaddr.sin_port);
    if (iport != oport) {
	sprintf(msg, "Port mapping: %d --> %d\n", 
	  iport&0xffff, oport&0xffff);
	sk_log(LOG_NOTICE, "%s", msg);
    }
    if (om&OLAUNCH_v) sk_log(LOG_DEBUG,
	    "Socket #%d on machine %s, port %d (stream i/o)\n",
	    sock, inet_ntoa(sockaddr.sin_addr), routine_no&0xffff);
    if (om&OLAUNCH_vv) sk_iolog(sk_getlog());

    ocalls[sock] = om | OLAUNCHED;		/* Save Mode */
    return(sock);
}

int sk_astart(service, mode, authfile)
/*++++++++++++++++
.PURPOSE  Start a TCP Server with an authorisation file
.RETURNS  The socket of file descriptor number / -1 if error
.REMARKS  Mode is defined in sk_start.
-----------------*/
  char *service;	/* IN: Service or Port Number */
  char *mode;		/* IN: Mode: d=Daemon,  v=Verbose l=SysLogging */
  char *authfile;	/* IN: Authentification file, or NULL(no check)\
				or "" (standard /etc/passwd file)*/
{
  int len;
    if (afile) free(afile), afile = (char *)0;
    if (authfile) {
    	len = 1 + strlen(authfile); afile = malloc(len); 
    	strcpy(afile, authfile);
    	sk_setauth ( afile ? passwd : (iFCT)0 );
    }
    return (sk_start(service, mode));
}

/*============================================================================*/

int sk_listen(sock, queue_slots)
/*++++++++++++++++
.PURPOSE  Listen to a socket started by sk_start (not in "inetd" mode),
	until a connection happens.
.RETURNS  The connection number / -1 if error
-----------------*/
  int sock;		/* IN: Started Socket Number 	*/
  int queue_slots;	/* IN: Maximal number of pending connections */
{
  int plug;
    plug = sk_accept(sock, queue_slots) ;
    if ((plug >= 0) && (sk_ack(plug) >= 0)) 
	return(plug);
    return(-1);
}

int sk_accept(sock, queue_slots)
/*++++++++++++++++
.PURPOSE  Listen to a socket started by sk_start (not in inetd mode),
	until a connection happens.
.RETURNS  The connection number / -1 if error
-----------------*/
  int sock;		/* IN: Started Socket Number 	*/
  int queue_slots;	/* IN: Maximal number of pending connections */
{
  int plug;				/* fd for the working socket */
  struct sockaddr_in caller;		/* id of foreign calling process */
  int fromlen = sizeof(caller);
  char om;
  int stat;

    om = ocalls[sock];	/* Options */
    if (om&OLAUNCH_d) {			/* Daemon */
    	if (om&OLAUNCHED) {			/* First connection */
	    if (getpeername(0, (struct sockaddr *)&caller, &fromlen) < 0) {
	    	sk_perror("Can't getpeername");
	    	return(-1);
	    }
	    if (om&OLAUNCH_v) sk_log (LOG_DEBUG, "getpeername succeeded\n");  
	}
	plug = 0;
    }
    else {			  /* Not a Daemon */
    	if (om&OLAUNCHED) {	/* First connection */
 	    /* put an ear to the socket, listening for a knock-knock-knocking */
  	    if (listen( sock, queue_slots > 0 ? queue_slots : 1 ) < 0) {
	    	sk_perror("****sk_listen/listen");
	    	return(-1);
	    }
	}
	ocalls[sock] &= ~OLAUNCHED;		/* Initilisation done */
    	    /* plug will be a socket different from sock */
	while(1) {
  	    plug = accept( sock, (struct sockaddr *)&caller, &fromlen);
	    if (plug >= 0) break;
	    if (errno == EINTR) continue;
	    sk_perror("****sk_listen/accept");
	    return(-1);
	}
    }
    if (plug >= NSOCKS) {
	sk_log(LOG_ERR, "Too many opened sockets (%d)\n", plug);
	close(plug);
	return (-1);
    }
    ocalls[sock] &= ~OLAUNCHED;		/* Initilisation done */
    ocalls[plug]  = ocalls[sock];	/* Inheritance */

    if (calls[plug]) free(calls[plug]);
    calls[plug] = (char *)0;
    theplug = plug;	/* For passwd & Signal usage */

    if (ncalls[plug]) free(ncalls[plug]);
    ncalls[plug] = strdup(inet_ntoa(caller.sin_addr));
    
    	/* print calling process' identification */
    if (om&OLAUNCH_v)sk_log(LOG_DEBUG, "Plug #%d from %s, port %d\n", 
	plug, ncalls[plug], ntohs(caller.sin_port));
#if 0
    if (om&OLAUNCH_v)sk_log(LOG_DEBUG, "Plug #%d, struct sockaddr_in {\
        sin_family      = %d\
        sin_addr.s_addr = %s\
        sin_port        = %d }\n", 
	plug, 
	caller.sin_family,
	inet_ntoa(caller.sin_addr),
	ntohs(caller.sin_port)
    );
#endif
#ifdef NODELAY
	stat = TCP_NODELAY;
	if (setsockopt(plug, IPPROTO_TCP, TCP_NODELAY, 
	    (char *)&stat,sizeof(stat))<0){
	    sk_perror("Can't setsockopt(TCP_NODELAY)");
	    close(plug);
	    return(-1);
	}
#endif
    sk_errfct(sk_perror); /* Replace Error Logging Function in skio routines */

    return(plug);
}

int sk_ack(plug)
/*++++++++++++++++
.PURPOSE  Acknowledgements with the client:
          Send Host/Service:pid.plug
          Receive Acknowledgment as User@Node:Remote-User
          Eventually, receive the userid/passwd and check them
.REMARK   MUST BE CALLED BY THE CHILD SERVER AFTER THE FORK
.RETURNS  The connection number (plug) / -1 if error
-----------------*/
  int plug;             /* IN: The connection number provided by sk_accept */
{
  int fromlen;
  char om, buf[120], client[80], *pc, *pu;
  int stat;

    om = ocalls[plug];  /* options */

	/* Send Host/Service:pid.plug and wait for Answer */
    gethostname(buf, sizeof(buf));
    sprintf( buf + strlen(buf), "/%s=%d:%d.%d", routine, routine_no,
    	getpid(), plug);
    if (om&OLAUNCH_v) sk_log(LOG_DEBUG, "Sending  <%s>\n", buf);
    fromlen = sk_puts ( plug, buf);
    if ( fromlen < 1 ) {
	sk_log(LOG_ERR, "Sent <%s> returned %d\n", buf, fromlen);
	close(plug);
	return(-1);
    }

    	/* read Acknowledgment as User@Node:Remote-User */
    fromlen = sk_gets(plug, client, sizeof(client));
    if ( fromlen < 1 ) {
	sk_log(LOG_ERR, "Acknowledgment of %d bytes <%s>\n", fromlen, client);
	close(plug);
	return(-1);
    }
    client[fromlen] = 0;
    client[--fromlen] = 0;		/* Replace the \n */
    if (om&OLAUNCH_v) sk_log(LOG_DEBUG, "Received <%s>\n", client);

	/* Store User@Internet(HostName):remote-name */
    for (pc = client; *pc; pc++)  if (*pc == '@') break;
    if (! *pc) {	/* ANSWER not as expected ! */
	sk_log(LOG_ERR, "Acknowledgment User@Internet(HostName) is:\n%s\n", 
		client);
    	sk_errsend(plug, "****Invalid Acknowledgment (Sorry)");
    	close(plug); 
    	return(-1);
    }
    *(pc++) = 0;
    for (pu = pc; *pu; pu++)  if (*pu == ':') break;
    if (*pu) *(pu++) = 0;
    sprintf(buf, "%s@%s(%s):%s", 
	client, ncalls[plug], pc, pu);
	/* ntohs(caller.sin_port)); */
    if (om&OLAUNCH_v) sk_log(LOG_DEBUG, "Caller = %s\n", buf);
    fromlen = strlen(buf)+1;
    calls[plug] = malloc(fromlen); strcpy(calls[plug], buf);

	/* If there is an Authentification Routine, ask for a Password */
    if (*pu || authentification) {	/* Password Sent or Required   */
    	sk_put(plug, "\06");		/* ^F means: Send Back Counted */
	stat = sk_getl(plug);
	if (stat > 0)  stat = sk_get(plug, buf, stat);
	if (stat >= 0) {
    	    buf[stat] = 0;		/* Terminate string */
    	    if (!authentification) stat = 0;
    	    else stat = (*authentification)(calls[plug], buf);
    	}
    	if (stat < 0) {
    	    sk_errsend(plug, "****Invalid Authentication (Sorry)");
    	    close(plug); 
    	    return(-1);
    	}
    }

	/* Send back OK = ^D */

    if (om&OLAUNCH_v) sk_log(LOG_DEBUG, "sk_listen OK, send ^D\n");
    sk_put(plug, "\04");

    return(plug);
}

/*============================================================================*/

#ifndef BASIC		/* TAPE i/o is not always required... */
static int iostat(msg)
/*++++++++++++++++
.PURPOSE  Check the tape status
.RETURNS  Status. msg is logged if necessary
-----------------*/
  char *msg;	/* IN: Message to print in case of error */
{
  int stat;

    stat = ioctl(thefile,MTIOCGET,&mt_stat);
#ifdef DEBUG
    printf("----Status mt: type=%X, stat=%X, err=%X, rbytes=%d\n",
	mt_stat.mt_type, mt_stat.mt_dsreg, mt_stat.mt_erreg, mt_stat.mt_resid);
#endif
    if (stat == -1) 
	sk_perror  ("Tape status"),
	sk_perrsend(theplug, "Tape status");
    else  {
	stat = mt_stat.mt_erreg & 037;
	if (stat) { char *b;
	    b = malloc(256+strlen(msg)); sprintf(b, 
		"****%s: mt_type=%4X, mt_dsreg=%4X, mt_erreg=%4X, mt_resid=%d",
	    	"mtget struct in mtio.h", 
		mt_stat.mt_type, mt_stat.mt_dsreg, mt_stat.mt_erreg, 
		mt_stat.mt_resid);
	    sk_perror  (b),
	    sk_perrsend(theplug, b);
	    free(b);
	}
    }
    return(stat);
}
#endif

int sk_servlog(opt)
/*++++++++++++++++
.PURPOSE  Set server log
.RETURNS  0 (opened) / not-zero (opened)
-----------------*/
  int opt;	/* IN: 1 to start/restart log   */
{
    if (opt) {
	deblog = fopen("/tmp/sk.log", "a");
	if (deblog) fprintf(deblog, "\n====> Open Debug Log <====\n");
	sk_iolog(deblog);
    } else if (deblog) {
	fclose(deblog);
	deblog = (FILE *)0;
    }
    return(deblog != (FILE *)0);
}

int sk_maxbytes(bytes)
/*++++++++++++++++
.PURPOSE  Install the max number of bytes for next transfer
.RETURNS  Previous value of maxbytes.
.REMARKS  Use sk_maxbytes(-1) to get the current value without change.
	Value 0 = Unlimited.
-----------------*/
  int bytes;	/* IN: Maximal value of number of bytes */
{
  int ret;
    ret = maxbytes;
    if (bytes >= 0) maxbytes = bytes;
    return(ret);
}

int sk_fromclient(plug, fh)
/*++++++++++++++++
.PURPOSE  Copy what comes from client to file.
.RETURNS  Number of bytes transferred (if 2^31-1, use sk_blocks to get
		block counts
-----------------*/
  int plug;	/* IN: Started Socket Number 	*/
  int fh;	/* IN: OUTput file handle	*/
{

  int lb, stat, bytes, got;
  char *b, *errmsg;
#ifndef BASIC
  struct mtop top;
  char mtx;		/* Mag-Tape Status */
#endif
#ifdef FROMCLIENT	/* Append lengths */
  int lbo, lbt;
  time_t tim; char *ct, *ctime();
  FILE *flog;
    got = lbo = lbt = 0;
    flog = fopen("/tmp/fromclient", "a");
    if (!flog) flog = stderr;
    time(&tim); ct = ctime(&tim); ct[20] = 0; 
    fprintf(flog, "====%s====%s[%d] (Start)", ct+4, "fromclient", getpid());
#endif

    theplug = plug; thefile = fh;
#ifndef BASIC	/* Check if a tape --- there are frequently problems !! */
    { struct stat st;
	fstat(fh, &st);
	if (st.st_mode & S_IFCHR) {
	    mtx = 1;
	    iostat("Starting Tape");
	    sk_log(LOG_INFO, "sk_fromclient: output is a TAPE\n");
	}
	else mtx = 0;
    }
#endif

    /* WRITE a file, i.e. READ the socket.
	The Control-F tells the client
	"Please send a Buffer Length+Contents" 
    */

    if (sk_write(plug, "\06", 1) != 1) return(-1);
    blocks = bytes = stat = lb = 0; b = errmsg = (char *)0;

    while (1) {
	stat = sk_getl(plug);
#ifdef FROMCLIENT
	if (stat != lbo) {
	    fprintf(flog, "/%d[%d]\n %5d/", got, lbt, stat);
	    lbo = stat; lbt = 0;
	}
#endif
	if (stat <= 0) break;
	if (lb < stat) { 
	    lb = (stat+1023)&(~1023); 	/* Round to next 1k */
	    if(b) free(b);
	    b = malloc(lb); 
	}
	got = sk_get(plug, b, stat);
#ifdef FROMCLIENT
	lbt++;
#endif
	/* If an error was encountered, edit it NOW, and stop */
	if (errmsg) {
	    sk_errsend(plug, errmsg) ;
	    if (*errmsg == '*') stat = -1;
	    break ;
	}
	if (got != stat) { stat = -1; break; }
	if (sk_write(plug, "\06", 1) != 1)  { stat = -1; break; }
	stat = write(fh, b, got);
	if (got != stat) { 	/* Could not write Device, e.g. Over 2Gbytes */
	  char buf[64], *p ;
	    sprintf(buf, "****Server Write File(%d bytes) returned %d -", 
		got, stat);
	    sk_perror(buf); 
	    if (stat < 0) p = strerror(errno) ;
	    else p = "" ;
	    errmsg = malloc(strlen(buf) + strlen(p) + 2) ;
	    strcpy(errmsg, buf); strcat(errmsg, p) ;
	    /* Cannot send the error here, because the ^F was just sent,
	       asking the client to send the next buffer .
	       That's what we continue here .
	    */
	    continue ;
	    /* sk_perrsend(plug, buf);  */
	    /*sk_log(LOG_ERR, "Server Write(%d bytes) returned %d\n", got, stat);
	    */
#ifndef BASIC
	    if (mtx) iostat("****Writing Tape****");
#endif
	    /* stat = -1; */
	    break; 
	}
	bytes += stat;
	if (maxbytes && (bytes > maxbytes)) {
	    errmsg = malloc(100);
	    sprintf(errmsg, "++++Transfer>%dbytes, truncated to: %d\n", 
	      maxbytes, bytes);
	    sk_log(LOG_WARNING, errmsg+4);
	}
	if (bytes >= MAXBYTES) {	/* Close to 2GBytes */
	   blocks += (bytes >> 9) ;
	   bytes -= MAXBYTES ;
	}
#ifndef BASIC
	if (mtx) {	/* Have to check evey blockl written on Tape */
#ifdef MTCSE
	    if (iostat("++++Clear serious exception++++")) {
		mtop.mt_op = MTCSE, mtop.mt_count = 1;
		ioctl(fh,MTIOCTOP,&mtop);
	    }
#else
	    if (iostat("Writing Tape")) stat = -1;
#endif
	}
#endif
    }
#ifdef DEBUG
    puts("");
#endif
    if (b) free(b);
#ifdef FROMCLIENT
    time(&tim); ct = ctime(&tim); ct[20] = 0; 
    fprintf(flog, "/%d[%d]", got, lbt);
    fprintf(flog, "\n====%s====%s[%d] (End)\n", ct+4, "fromclient", getpid());
    fclose(flog);
#endif
    if (errmsg) free(errmsg);
    if (stat < 0) return(stat) ;
    if (blocks == 0) return(bytes) ;
    if ((blocks == (MAXBYTES>>9)) && (bytes <= 0xffff))
	return(MAXBYTES+bytes) ;
    return(0x7fffffff) ;
}

int sk_toclient(plug, fh)
/*++++++++++++++++
.PURPOSE  Copy opened file to client
.RETURNS  Number of bytes transferred
.REMARKS  Can be interrupted by the Client (^C)
-----------------*/
  int plug;	/* IN: Started Socket Number 	*/
  int fh;	/* IN: INput file handle	*/
{
  int lb, stat, bytes;
  char *b, *errmsg;

    /* READ a file, i.e. WRITE to socket.
	The Control-B tells the client
	"From now, you'll get data in Buffered Mode" 
    */
    if (sk_write(plug, "\02", 1) != 1) return(-1);
    blocks = bytes = stat = 0; errmsg = (char *)0;
    lb = 60*1024; b = malloc(lb);
    stat = sk_read(plug, b, 1);	/* Read Ack */
    if (stat < 0) goto Fin;

    while (1) {
	stat = read(fh, b, lb); 
	if (deblog) {
	    fprintf(deblog, "sk_toclient: read %d bytes", stat);
	    if (stat > 0) fprintf(deblog, " <"),
		fwrite(b, 1, stat, deblog), 
		fprintf(deblog, ">");
	    fprintf(deblog, "\n");
	    fflush(deblog);
	}
	if (stat < 0) { 
	    sk_perror("****Server Reading File"); 
	    sk_putl(plug, 0);	/* Indicates End */
	    sk_perrsend(plug, "****Server Reading File"); 
	    break; 
	}
	if (sk_putl(plug, stat) < 0) { stat = -1; break; }
	if (stat == 0) break; 	/* END indicator */
	if (errmsg) {
	    sk_errsend(plug, errmsg) ;
	    stat = -1;
	    break ;
	}
	if (sk_write(plug, b, stat) != stat) { stat = -1; break; }
	bytes += stat;
	if (maxbytes && (bytes > maxbytes)) {
	    errmsg = malloc(100);
	    sprintf(errmsg, "Transfer>%d, truncated to bytes: %d", 
	      maxbytes, bytes);
	}
	if (bytes >= MAXBYTES) {	/* Close to 2GBytes */
	   blocks += (bytes >> 9) ;
	   bytes -= MAXBYTES ;
	}
    }
  Fin:
    if (b) free(b);
    if (errmsg) free(errmsg);
    /* Return: limit to 2^31-1 bytes */
    if (stat < 0) return(stat) ;
    if (blocks == 0) return(bytes) ;
    if ((blocks == (MAXBYTES>>9)) && (bytes <= 0xffff))
	return(MAXBYTES+bytes) ;
    return(0x7fffffff) ;
}
